'''
Created on 20 févr. 2019

@author: jeanluc
'''
import os
import numpy as np
from pandas import DataFrame
from string import ascii_uppercase
#from random import choice
import random as rnd
from pandas import ExcelWriter

#import metaphor
from ..nntoolbox.datareader.datasource import DataSource
from ..nntoolbox.utils import orderedIndexes, make_extern
from ..nntoolbox.lineartoolbox import randomProbe  
from ..nntoolbox.utils import doNothing, setRandSeed
from ..nntoolbox.cmd_line_dialogs import choiceQuestion, \
    valueQuestion, yesNoQuestion #, chooseFields
from ..nntoolbox.progress import Progress_tqdm  #FanWait, 
from ..nntoolbox.constants import nobarverb
from ..nntoolbox.excelutils import enrichExcelPreproc
from ..nntoolbox.filetoolbox import decorfile

from ..nn1.api._api_service import CloseExcelwriter  #, getExcelFileName, createExcelwriter 
#from 

normalizeVectors = False # Ne sert a rien de mettre "True"

preprocdefaults = {
    'dopreproc': 1,
    'new_threshold': 0,
    'new_preproc': 0,
    }

dialogdefaults = preprocdefaults
preproctypelist = ['', 'linear', 'crossproduct', 'full2ndorder']
""" crossproduct -> only cross products in the second order.
    full2ndorder -> full second order.
"""
preprocnamelist = ['', 'linear', 'linear + cross-products', 'full quadratic']
    
def printresult(options, message, vecind, Pertinence, inputTitles, choices=None, 
        printTitles=True, resultstr="score", doprint=doNothing):
    doprint(message)
    lmax = max([len(val) for val in inputTitles])
    nametitle = "input{0}".format(" "*(lmax-4))
    if printTitles:
        doprint("\t".join(["ind", nametitle, resultstr]))
    addon = [lmax-len(val) for val in inputTitles]
    for ind in vecind:
        if choices is not None:
            choicex = choices[ind]
        else:
            choicex = True
        if choicex:
            prob = "{0:<4.3g}".format(Pertinence[ind])
            supp = " " * addon[ind]
            title = "{0}{1}".format(inputTitles[ind], supp)
            doprint(ind, title, prob, sep="\t")        

def doPreProc(options, Inputs, Outputs, inputTitles, probecount=-1, 
        randomSource='normal', doprint=doNothing):
    
    if not options.preproctype in preproctypelist[1:]:
        doprint("Random probe analysis failed")
        return None, None
    
    preprocindex = preproctypelist[1:].index(options.preproctype)
    directsquare = preprocindex == 2 
    crossprod = preprocindex > 0
    testconstitution = (crossprod, directsquare)
    if probecount < 0:
        mess = "Enter number of random probe tests"
        probecount = valueQuestion(mess, options.probecount, 
            "Probe tests", doprint=doprint, verbose=options.verbose, 
            forcedefault=options.yes)  #, prefix=prefix)
    options.probecount = probecount
    
    with Progress_tqdm(options.probecount, desc=options.preproctype,   #'random probe', 
            nobar=options.verbose<nobarverb) as update:
        
        ndc, indexes, results2 = randomProbe(Inputs.T, Outputs.T, 
            options.probecount, seed=options.seed, randomSource=randomSource, 
            testconstitution=testconstitution, progress=update,
            detailedoutput=True, donorm=normalizeVectors) 
        
        update.ends()

    Pertinence = np.zeros((ndc))
    for index, value in zip(indexes, results2):
        for ind in index:
            Pertinence[ind] = max(Pertinence[ind], value)  
    vecind = orderedIndexes(Pertinence)
    message = "Probe results with {0} selection ({1}):".format(options.preproctype, options.probecount)
    printresult(options, message, vecind, Pertinence, 
        inputTitles=inputTitles, printTitles=True, doprint=doprint)
    return Pertinence, vecind, indexes, results2

def _preprocAction(options, doprint=doNothing):
    
    preprocSheetName = "Pre Processing"
    seedmem = options.seed
    pertinenceFrame, vecinds, localframes = _doPreProcToFrames(options, doprint=doprint)
    
    if options.copyexcel: 
        tempstr = ''.join([rnd.choice(ascii_uppercase) for _ in range(8)])
        fname = os.path.join(options.targetfolder, "{0}.xlsx".format(tempstr))
        excelwriter = ExcelWriter(fname, engine='xlsxwriter')
        pertinenceFrame.to_excel(excelwriter, preprocSheetName)
        for key, frame in localframes.items():
            frame.to_excel(excelwriter, key)
        for key in localframes.keys():
            enrichExcelPreproc(excelwriter, key, fname)
        enrichExcelPreproc(excelwriter, preprocSheetName, fname, True)
        fname = CloseExcelwriter(options, excelwriter, forcename="PreProc", 
            forbiden=['hidden'], Sheet=preprocSheetName)
    elif options.copytxtcsv and os.path.exists(options.targetfolder):
        fname = os.path.join(options.targetfolder, 'resultFrame')
        fname = decorfile(fname, options, ".{0}".format(options.copytype), indextemplate="(%d)")                    
        with open(fname, 'w') as ff:
            ff.write(pertinenceFrame.to_csv(sep=options.sep, decimal=options.decimal))
    options.seed = seedmem
    externpath = make_extern(fname, options)
    doprint("Pertinence analysis is saved in file : {}".format(externpath))
    inputTitles = list(pertinenceFrame.index)
    dopreproc = options.preproctype in preproctypelist[1:] 
    res = False 
    while dopreproc:
        preproctype = preproctypelist.index(options.preproctype)  
        mess = "Choose the preprocessor action by its index."
        printMessage = "Preprocessor action"
        choicelist = ["No_Action", "Linear random probe", "Linear + cross-product random probe", "Full 2nd order random probe"]
#         try:
        preproctype, index = choiceQuestion(mess, choicelist, 
            preproctype, printMessage, returnType=str, doprint=doprint, 
            extradisplay=True, listindexed=True, forcedefault=options.yes,
            fulloutput=True)
#         except QuestionInterrupt:
#             
        if not index: 
            return None
        options.preproctype = preproctypelist[index]
        Pertinence = pertinenceFrame.ix[index-1]
        vecind = vecinds[index-1]
        
        if Pertinence is None:
            return -1
        doselect = True
        while doselect:
            try:
                options.pertinencethreshold = valueQuestion("Enter the input selection pertinence", 
                    options.pertinencethreshold, "Input selection pertinence", 
                    forcedefault=options.yes, doprint=doprint, 
                    verbose=options.verbose, returnType=float) 
            except Exception as err:
                doprint(str(err))
                return 1
            res = False            
            if (0.0 < options.pertinencethreshold < 1.0):
                PertColName = pertinenceFrame.columns[index-1]
                Perts = list(pertinenceFrame[PertColName])
                choices = [val < options.pertinencethreshold for val in Perts]
                if any(choices):
                    mess = "Following inputs may be removed:"
                    printresult(options, mess, vecind, Perts, 
                        inputTitles=inputTitles, choices=choices, printTitles=False, 
                        doprint=doprint)
                    mess = "Do you accept?"
                    selcount = 0
                    for choicex in choices:
                        if not choicex:
                            selcount += 1
                    OKChoice = vecind[:selcount]
                    OKChoice.sort()
                    OKTitles = [inputTitles[val] for val in OKChoice]
                    st = ", ".join(OKTitles)
                    printMessage = ""  
                    res = yesNoQuestion(mess, default='Y', printMessage=printMessage, 
                        forcedefault=options.yes, verbose=options.verbose, 
                        doprint=doprint)
                    if res:
                        inputindex = options.grouplist.index('inputs')
                        inputini = options.datafields[inputindex]
                        
                        Inputs = [val for ind, val in enumerate(inputini) if ind in OKChoice]
                        options.groupcount[inputindex] = len(Inputs)
                        options.datafields[inputindex] = Inputs
                        if options.test:
                            testinputini = options.testfields[inputindex]
                            Inputs = [val for ind, val in enumerate(testinputini) if ind in OKChoice]
                            options.testfields[inputindex] = Inputs
                if res:
                    doselect = False
                else:
                    mess = "will you try a new selection threshold ? "
                    valdef = int(dialogdefaults['preproc']['new_threshold'])
                    doselect = yesNoQuestion(mess, default=valdef, printMessage="", 
                            forcedefault=options.yes, verbose=options.verbose, 
                            doprint=doprint)
        if res:
            dopreproc = False
            doprint("Inputs will be {0}".format(st))
        else:
            mess = "will you try a new pre-processing ? "
            valdef = int(dialogdefaults['preproc']['new_preproc'])
            dopreproc = yesNoQuestion(mess, default='y', printMessage="", 
                    forcedefault=options.yes, verbose=options.verbose, 
                    doprint=doprint)
    if res:
        return options
    return None


def _doPreProcToFrames(options, verbose=3, doprint=doNothing):
    
    options.seed = valueQuestion("Enter a random seed (-1 for no seed)", 
        options.seed, "Seed",  
        doprint=doprint, verbose=verbose, returnType=int)  #forcedefault=forceyes, prefix=prefix,
    setRandSeed(options.seed)
    curpreproc = options.preproctype
    source = DataSource(options.source, options.filetype, options.datarange, 
        datafmt=options.datafields, titles=True)
    datas, titles = source.getDataArrays(options.datarange, 
        fieldsList=options.datafields, titles=True)  
    inputTitles = titles[options.grouplist.index('inputs')]
    Inputs = np.asarray(datas[options.grouplist.index('inputs')])
    Outputs = np.asarray(datas[options.grouplist.index('outputs')])
    perts = []
    vecinds = []
    options.probecount = valueQuestion("Enter number of random probe tests", 
        options.probecount, "Probe tests", doprint=doprint, 
        verbose=options.verbose, forcedefault=options.yes)  

    choicelist = ['normal', 'shuffle']
    printMessage = "Random method"
    mess = "Please choose the random method :"
    randomSource, _ = choiceQuestion(mess, choicelist, defaultindex=1, 
        printMessage=printMessage, returnType=int, extradisplay=True, 
        listindexed=True, fulloutput=True, doprint=doprint)
    options.randomSource = randomSource
    localframes = {}
    for preproctype in preproctypelist[1:]:
        options.preproctype = preproctype
        Pertinence, vecind, indexes, results2 = doPreProc(options, Inputs, 
                                                          Outputs, 
            inputTitles, randomSource=randomSource,
            probecount=options.probecount) 
        if Pertinence is None:
            break
        pertlist = list(Pertinence)
        Pertinence = np.asarray(pertlist)
        perts.append(Pertinence)
        vecinds.append(vecind)
        localTitles = [(" * ".join([inputTitles[val] for val in indx])) 
                       for indx in indexes]
        localframe = DataFrame(results2, columns=[preproctype], 
                               index=localTitles)
        localframes[str(preproctype)] = localframe 
    ll = list(map(list, zip(*perts)))
    pertinenceFrame = DataFrame(ll, columns=preprocnamelist[1:], 
                                index=inputTitles) 
    
    if doprint:
        doprint(pertinenceFrame)
    
    options.preproctype = curpreproc
    return pertinenceFrame, vecinds, localframes#, randomSource

if __name__ == '__main__':
    pass


#-*- coding: ISO-8859-15 -*-
#===============================================================================
# $Id: specialmath.py 4789 2018-06-08 11:40:08Z jeanluc $
#  Module  _specialmath
#  Projet MonalPy
# 
#    Ce module pourra etre supprime avec python 3.x
# 
#  Implementation python de monal
#
#
# Copyright 2016 Jean-Luc PLOIX
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import sys, os  #, math
from math import  sqrt, isinf, isnan, fabs, fsum
from numpy.random import random
import numpy as np

try:
    from scipy.linalg import svd
except ImportError: pass

from .monalconst import LEV_EARLY, LEV_MONARI

assert sys.version_info > (2, 6) 

'''
Created on 30 déc. 2009

@author: Jean-Luc PLOIX
'''
DBL_MAX = 1.79769313486231e+308
DBL_MIN = 2.2250738585072013e-308

def sqr(X):
    return X*X
    
def isinfinite(value):
    return isinf(value)

def ispinfinite(value):
    return isinf(value) and (value > 0)

def isminfinite(value):
    return isinf(value) and (value < 0)

def isspecial(value):
    return isnan(value) or isinf(value)

class _Special(float):
    """_Special numbers: Inf, -Inf, NaN"""
    def __init__(self): super(_Special, self).__init__()
    def __int__(self): return self
    def __float__(self): return self
    def __repr__(self): return "%s()"% self.__class__.__name__
    def __eq__(self, val): return False
    def __ne__(self, val):  return not isinstance(val, self.__class__)
    def __radd__(self, val): return self.__add__(val)
    def __rmul__(self, val): return self.__mul__(val)
    def __bool__(self): return True
    def __hash__(self): return hash(self.__repr__())
        
class _baseInfinity(_Special):
    """Tous les infinis"""
    def otherimplementation(self): return None
    def __abs__(self): return INF
    def __mul__(self, val): 
        if isinstance(val, _NaN) or (val == 0): return NaN
        if val < 0: return self.otherimplementation()
        return self
    def __neg__(self): return self.otherimplementation()
    def __sub__(self, val): 
        if isinstance(val, (self.__class__, _NaN)): return NaN
        return self
    def __rsub__(self, val): 
        if isinstance(val, (self.__class__, _NaN)): return NaN
        return _NegInfinity()
    def __div__(self, val): 
        if isspecial(val): return NaN
        return self
    def __rdiv__(self, val): 
        if isspecial(val): return NaN
        return 0.0
    def __invert__(self):
        return 0.0

class _Infinity(_baseInfinity):
    """Classe représentant l'infini positif
"""
    def otherimplementation(self): return NEGINF
    def __str__(self): return 'INF'
    def sqrt(self, context=None): return self
    def __add__(self, val):
        if isinstance(val, (_NegInfinity, _NaN)): return NaN
        return self
    def __gt__(self, val):
        if isinstance(val, self.__class__): return False
        return True
    def __ge__(self, val):
        if isinstance(val, self.__class__): return False
        return True
    
class _NegInfinity(_baseInfinity):
    """Classe représentant l'infini negatif
"""
    def otherimplementation(self): return INF
    def __str__(self): return "-INF"
    def __add__(self, val): 
        if isinstance(val, (_Infinity, _NaN)): return NaN
        return self
    def __lt__(self, val):
        if isinstance(val, self.__class__): return False
        return True
    def __le__(self, val):
        if isinstance(val, self.__class__): return False
        return True
    
class _NaN(_Special):
    """Classe  représentant l'indetermination (le résultat de la division 0/0).
"""
    def __str__(self): return "NaN"
    def __rsub__(self, val): return self
    def __gt__(self, val): return False
    def __ge__(self, val): return False
    def __neg__(self): return self
    
INF = infinity = float('inf')
NEGINF = neginfinity = float('-inf')
NAN = nan = NaN = float('nan')

def dotex(*argv):
    # extension de la fonction numpy.dot a un nb qcq de matrices
    res = None
    for val in argv:
        if res is None:
            res = val
        else:
            res = np.dot(res, val)
    return res

def leverage(vec, mat):
    # compute zT M z
    return dotex(vec.T, mat, vec) 

def getCorrelMatrix(source, eye=1):
    # source est une matrice carrée symetrique
    """Creation de la matrice de correlation entre les paramètres à partir de la matrice de dispersion.
    La diagonale est mise à la valeur "eye"(1 par défaut).
    """ 
    N = source.shape[0]
    target = np.zeros((N,N))
    diag = np.diag(source)
    for i in range(N):
        target[i, i] = eye
        for j in range(i):
            target[i, j] = target[j, i] = source[i, j]/ sqrt(diag[i]*diag[j])
    return target

def getCorrelFromDispersion(dispersion):
    q, q2 = dispersion.shape
    assert q == q2
    correl =  np.zeros((q, q))
    diag = np.diag(dispersion)

    for i in range(q):
        for j in range(i):
            coeffa = diag[i]*diag[j]
            try:
                coeff = sqrt(coeffa)
            except ValueError:
                coeff = sqrt(-coeffa)
            correl[i, j] = correl[j, i] = dispersion[i, j]/ coeff
    return correl
    
# def getSVDChildrenOld(jacob):
#     """Calcul et exploitation de la decomposition SVD d'une matrice Jacobienne.
#     Retour :
#         matrice de dispersion 1/(ZtZ)
#         vecteur des leviers
#         matrice de correlation avec la diagonale à 0
#         determinant de la matrice de dispersion
#         vecteur des valeurs propres de la matrice de dispersion
#     """
#     N, q = jacob.shape
#     U, s, V = svd(jacob, full_matrices=False)
#     eigs = s**-2
#     dispersionMatrix = dot(dot(V.T, npdiag(eigs)), V)
#     leverages = zeros((N,))
#     correlMat = getCorrelFromDispersion(dispersionMatrix)
#     for i, Ui in enumerate(U):
#         Ui2 = Ui**2
#         Ui2.sort()
#         leverages[i] = sum(Ui2)
#     determinant = prod(s)
#     return dispersionMatrix, leverages, correlMat, determinant, eigs

def getLeveragesFromSVD(jacob):
    N, q = jacob.shape
    try:
        U, s, V = svd(jacob, False)
    except Exception:
        raise
    projectionMat = np.dot(U, U.T)
    return np.diag(projectionMat)

def getSVDChildren(jacob, style=0,  method=LEV_EARLY):
    """Calcul et exploitation de la decomposition SVD d'une matrice Jacobienne.
    Retour :
        matrice de demidispersion 
        vecteur des leviers
        matrice de correlation avec la diagonale à 0
        determinant de la matrice de dispersion
        vecteur des valeurs propres de la matrice de dispersion
    """
    N, q = jacob.shape
    try:
        U, s, V = svd(jacob, False)
    except Exception:
        raise
    eigs = np.zeros((q,))
    #s**-2
    determinant = 1
    rank = q  
    #VV = V.copy()
    for i in range(q):
        if fabs(s[i]) > 1E-10:
            V[i] /= s[i]
            determinant *= s[i]
            eigs[i] = s[i]**2
        else:
            V[i] *= 0.0
            rank -= 1
            eigs[i] = NaN
    halfdispersion = V.T
    if style == 1:
        return halfdispersion
    dispersionMatrix = V.T.dot(V)
    correlMat = np.zeros(dispersionMatrix.shape)
    diag = np.diag(dispersionMatrix)
    for i in range(q):
        correlMat[i, i] = 0
        for j in range(i):
            correlMat[i, j] = correlMat[j, i] = dispersionMatrix[i, j]/ sqrt(diag[i]*diag[j])
    if method == LEV_EARLY:
        levs = np.asarray([sum(ui**2) for ui in U])
    elif method == LEV_MONARI:
        mat = jacob.dot(halfdispersion)
        levs = np.asarray([vec.dot(vec.T) for vec in mat])
    else:
        dispersion = dispersionMatrix  #dot(halfdispersion, halfdispersion.T)
        levs = np.asarray([vec.T.dot(dispersion.dot(vec)) for vec in jacob])
    if method in [LEV_EARLY]:  #, LEV_MONARI
        #dispersion = dispersionMatrix
        #levUs = asarray([vec.T.dot(dispersion.dot(vec)) for vec in jacob])
        mat = jacob.dot(halfdispersion)
        levUs = np.asarray([vec.dot(vec.T) for vec in mat])
    else:
        levUs = levs
    return halfdispersion, levs, levUs, correlMat, determinant, eigs, rank

def s_leverage2(grad, VW): 
    """Calcul du levier par la méthode détournée proposée par Gaétan Monari.
    W et V viennent du SVD du jacobien.
    W est le vecteur des valeurs singulières, par ordre decroissant
    V est la troisième matrice fournie par svd.
    """
    P, Q = VW.shape
    if grad.shape[0] != P:
        raise ValueError("matrices are not aligned")
    lst = []
    for k in range(P):
        lstk = (grdi*vec[k] for grdi, vec in zip(grad, VW))
        val = fsum(lstk)
        lst.append(val*val)
    return fsum(lst)
        
def s_leverage(grad, VW): 
    """Calcul du levier par la méthode détournée proposée par Gaétan Monari.
    W et V viennent du SVD du jacobien.
    W est le vecteur des valeurs singulières, par ordre decroissant
    V est la troisième matrice fournie par svd.
    """
    if grad.shape[0] != VW.shape[0]:
        raise ValueError("matrices are not aligned")
    lst = []
    for vec in VW:
        val = fsum(veci*gradi for veci, gradi in zip(vec, grad))
        lst.append(sqr(val))
#<<<<<<< .mine
    return fsum(lst)
    #return fsum(sqr(dot(vec, grad)) for vec in VW)
# =======
#     return math.fsum(lst)
#     #return math.fsum(sqr(dot(vec, grad)) for vec in VW)
# >>>>>>> .r4366
    
def leverages(jac):
    N, Q = jac.shape # Nex, Nparam
    U, W, V = svd(jac, False)
    VW = np.asarray([vec/W[ind] for ind, vec in enumerate(V)])
    lst = [s_leverage(grad, VW) for grad in jac]
    return np.asarray(lst)
        
def leverages2(jac):
    N, Q = jac.shape # Nex, Nparam
    U, W, V = svd(jac, False)
    res = np.zeros((N,))
    for k in range(N):
        p = Q-1
        while p>= 0:
            w = W[p]
            res[k] += (jac[k].dot(V[p])/w)**2
            p -= 1
    return res
        
if __name__ == '__main__':
    from pylab import plot, scatter, show
    
    jacalea = 0
    
    import os, sys
    from importlib import import_module
    
    dir = "C:\Projets\GM\EnergieLibreSolv\BaseA138" 
    module = "basea138_iso05n"
    filename = "DsolvG_basea138_iso05n_8.gmx"
    dir = "C:\Projets\GM\DSolvG\BaseA138"
    module = "basea138_iso05n"
    filename = "DsolvG_basea138_iso05n_1.gmx"
    dir = "C:\Projets\GM\LogKd109\Base109"
    module = "base109_chi05n"
    filename = "LogKd_base109_chi05n_17.gmx"
    sys.path.append(dir)
    os.chdir(dir)
    mod = import_module(module)
    
    if jacalea:
        jac = random((138,81))*2-1
    else:
        with np.load(filename, allow_pickle=True) as source:
            pp = source["params"]
        jac = mod.getjacobian(pp)
    
    lev4 = leverages(jac)
    print("lev4", lev4)
    print("sum lev4", sum(lev4))
    
    halfdispersion, projectionMat, correlMat, determinant, eigs, rank = getSVDChildren(jac)
    dispersionMatrix2 = halfdispersion.dot(halfdispersion.T)
    
    #print dispersionMatrix
    #print
    #print dispersionMatrix2
    
    U, s, Vh = svd(jac, False)
# <<<<<<< .mine
    #mat = U.dot(npdiag(s)).dot(Vh)
# =======
#     #mat = U.dot(diag(s)).dot(Vh)
# >>>>>>> .r4366
    #for a, b in zip(jac.flat, mat.flat):
    #    print a, b
    #Q, R = linalg.qr(jac, mode="economic")
    
# <<<<<<< .mine
#     #W2 = npdiag(s**-2)
# =======
#     #W2 = diag(s**-2)
# >>>>>>> .r4366
#     #M = Vh.T.dot(W2).dot(Vh)
    
    for i in range(s.shape[0]):
        Vh[i] /= s[i]
    lev3 = np.asarray([s_leverage(grd, Vh) for grd in jac])
    
    lev0 = np.diag(U.dot(U.T))
    print("Lev_U")
    print(lev0)
    print("sum Lev_U", sum(lev0))
    
    lev1 = np.asarray([s_leverage2(grd, halfdispersion.T) for grd in jac])
    print("Lev_G")
    print(lev1)
    print("sum Lev_G", sum(lev1))
    
    #===========================================================================
# <<<<<<< .mine
#     # lev2 = npdiag(Q.dot(Q.T))
# =======
#     # lev2 = diag(Q.dot(Q.T))
# >>>>>>> .r4366
    # if jacalea:
    #    print "Lev_Q"
    #    print lev2
    #    print "sum Lev_Q", sum(lev2)
    #===========================================================================
    
    #V = Vh.T
    #===========================================================================
    # Q = Vh.shape[0]
    # for i in xrange(Q):
    #   Vh[i] /= s[i]
    # V = Vh.T
    #===========================================================================
    #V = halfdispersion
    #s = scatter(lev3, lev1)
    #show()
    s1 = plot((0,1), (0,1))
    s1 = scatter(lev0, lev1)
    show()
    
    #print R.dot(R.T)
    #===========================================================================
# <<<<<<< .mine
#     # jac2 = random.random((138,81))*2-1
# =======
#     # jac2 = random((138,81))*2-1
# >>>>>>> .r4366
    # print jac2
    # print
    # print
# <<<<<<< .mine
#     # savez("jac2.npy")
# =======
#     # np .savez("jac2.npy")
# >>>>>>> .r4366
    # 
    # U, s, Vh = svd(jac2, False)
    # Q, R = linalg.qr(jac2, mode="economic")
    # 
# <<<<<<< .mine
#     # W2 = npdiag(s**-2)
# =======
#     # W2 = diag(s**-2)
# >>>>>>> .r4366
    # M = Vh.T.dot(W2).dot(Vh)
    # 
# <<<<<<< .mine
#     # lev0 = npdiag(U.dot(U.T))
# =======
#     # lev0 = diag(U.dot(U.T))
# >>>>>>> .r4366
    # print "lev_U"
    # print lev0
    # print "sum lev0", sum(lev0)
    # 
# <<<<<<< .mine
#     # lev2 = npdiag(Q.dot(Q.T))
# =======
#     # lev2 = diag(Q.dot(Q.T))
# >>>>>>> .r4366
    # print "lev_Q"
    # print lev2
    # print "sum lev2", sum(lev2)
    # 
    # #V = Vh.T
    # Q = Vh.shape[0]
    # for i in xrange(Q):
    #    Vh[i] /= s[i]
    # V = Vh.T
    # lev1 = asarray([s_leverage(grd, V, None) for grd in jac2])
    # print "lev_G"
    # print lev1
    # print "sum lev1", sum(lev1)
    #===========================================================================
# <<<<<<< .mine
#     #lev2 = npdiag(jac.dot(M).dot(jac.T))
# =======
#     #lev2 = diag(jac.dot(M).dot(jac.T))
# >>>>>>> .r4366
    #print lev2.shape
    
    #print "lev2"
    #print sum(lev2)
    #print lev2
    print("done")
    


#-*- coding: ISO-8859-15 -*-
#===============================================================================
# $Id: _driverlib.py 4565 2017-08-25 10:44:17Z jeanluc $
#  Module  _DriverLib
#  Projet MonalPy
# 
#  Implementation python de monal
# interface pour Dll
#
#  Author: Jean-Luc PLOIX  -  NETRAL
#  Mai 2014
#
# Copyright 2016 Jean-Luc PLOIX
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import os, sys
from six import PY3, text_type

from ctypes import CFUNCTYPE, byref, c_int, c_long, c_double, c_void_p, c_char_p
from weakref import proxy
import random
#from functools import reduce
from numpy import asarray, ndarray, zeros, ones, load, NaN, reshape, sqrt, \
    isnan, squeeze
from numpy import int as npint, long as nplong
from numpy import double as npdouble
from numpy import float as npsum
from numpy import sum as npflaot
from numpy import dot as npdot

from metaphor.nntoolbox.utils import floatEx, isNum, floatEx, decodequote, \
    RE_NONE, RE_LOO, RE_SUPER, BSGROUP, seedList, insertInList

from metaphor.nntoolbox.datareader.datasource import DataSource
from metaphor.nntoolbox.constants import USE_PARALLEL_MAIN, USE_PARALLEL_BS, \
    USE_SUPER_TRAIN, SORT_LOO_BS_PARALEL_TRAIN
from metaphor.nntoolbox.utils import concatenateList
from metaphor.monal.Property import Property #, Properties, delegate
from metaphor.monal.library.libmanager import LibManager, LibraryManagerError
from metaphor.monal import monalconst as C, specialmath as sp
from metaphor.monal.lcode import codeconst as CC
from metaphor.nntoolbox.utils import maxWorkers
from metaphor.monal.driver import DriverBase
from metaphor.monal.driver._parallelJobs import trainJobBS, trainJob, no_parallel_seed
from metaphor.monal.driver import optim_bfgs, optim_leastsq

import math
#from api.api import inpts
try:
    from scipy.stats.t import interval
    from scipy import optimize
    OKScipy = True
except ImportError:
    OKScipy = False
    from metaphor.monal.util.monaltoolbox import interval
try:
    import concurrent.futures as cf
except: pass

DEBUG_LOO = 0
DEBUG_LOO_FILE = 0
DEBUG_LOCAL_TRAIN = 0

def callbackminprint(cost, count):
    print("cost({0}): {1}".format(count, cost))
    # le return = 0 pour continuer. Une valeur True arrete l'operation
    return 0

def callbackmin(cost, count):
    return 0

# def callbackwrite(cost, count):
#     print("cost({0}) = {1}".format(count, cost))
#     return 0

class DriverLibError (Exception): 
    """DriverLib Exception.
    """
    pass

class DriverLib(DriverBase, LibManager):
    """class DriverLib
    Gestionnaire des drivers d'apprentissage multimodeles provenant de librairies dynamiques (dll, so, dylib)
    """
    
    def __init__(self, libfilename, verbose=0):  #, callback=None): # DriverLib
        DriverBase.__init__(self, verbose)  #, callback)
        LibManager.__init__(self, libfilename)
        try:
            self._modelType = self.moduleType
        except:
            self._modelType = 3
        self.savedir = ""
        self._targetfile = ""
        self._lastweights = None
        self._weights = None
        self.cumulEpoch = 0
        self.stoptrain = 0   
#        self.resultMatrix = None   
        self.presses = None
        self._callbacklib = None 
        self._callbacktype = None
        
        if self.propertyName == 'unknown':
            self.lib.setproperty(None)
#         if self._callback and self.lib:
#             self.setcallback(self._callback)      
    
    def __del__(self):
        LibManager.__del__(self)
        DriverBase.__del__(self)
    
    def reprlist(self, mode='long', prefix="module"):
        lst = ["{0} {1}:".format(prefix, self.moduleName)]
        if mode == 'long':
            lst.append("code module/program %s/%s"% (self.codeVersion, CC.__version__))
            lst.append("type %s"% self.moduleTypeStr)
            if self.mark:
                lst.append("mark: %s" % self.mark)
        if self.created and not self.mark:
            lst.append("created %s" % self.created)
        if self.base:
            lst.append("base: %s" % self.base)
        if self.propertyName:
            lst.append("property: %s" % self.propertyName)
        if self.inputCount:
            lst.append("inputs: {0}".format(self.inputCount))
        lst.append("hidden: %s"% self.hidden)
        lst.append("outputs: {0}".format(self.outputCount))
        lst.append("parameter count: %s" % self.paramCount)
        if mode == 'long':
            if self.biasbased is not None:
                lst.append("biasbased: %s"%  self.biasbased)
            if self.outlinks is not None:
                lst.append("outlinks: %s"%  self.outlinks)
            if self.targets is not None and len(self.targets):
                lst.append("outputnorm: %s"% self.outputnorm)
                if len(self.targets) <= 10:
                    tg = "targets %s"% self.targets
                else:
                    l1 = [str(val) for val in self.targets[:5]]
                    l2 = [str(val) for val in self.targets[-5:]]
                    tg = "targets [%s]"% " ".join(l1 + ["..."] + l2)
                lst.append(tg)
        return lst
    
    def repr(self, mode='long', tab="\t", prefix="module"):
        lst = self.reprlist(mode=mode, prefix=prefix)
        fulltab = "\n%s" % tab
        return fulltab.join(lst) 
    
    def __repr__(self):
        return self.repr('long')
    
    def xml(self): # LibManager
        st = self.getStrPropertyFromLib("xml")
        st = decodequote(st)
        return st
    
    def dict(self):
        return {
            'name': self.moduleName,
            'type': self.moduleType,
            'codeversion': self.codeVersion,
            'created': self.created,
            'mark': self.mark,
#            'config': self.configuration,
            'base': self.base,
            'property': self.propertyName,
            'paramCount': self.paramCount,
            'dataCount': self.dataCount,
            'hidden': self.hidden,
            }

    @Property
    def moduleTypeStr(self):
        if self.moduleType == 1:
            return 'usage, single'
        elif self.moduleType == 2:
            return 'training, multiple'
        elif self.moduleType == 3:
            return 'training, single'
        else:
            return 'unknown'
    
    @Property
    def paramCount(self):  # DriverLib
        return self.lib.paramCount()
        
    @Property
    def weights(self): # DriverLib
        return self._weights
    @weights.setter
    def weights(self, value):
        if isinstance(value, (list, ndarray)):
            self._weights = value.copy()
        else:
            self._weights = asarray(value, npdouble)
    
    def setWeights(self, weights):  # DriverLib
        self.weights = weights.copy()
    
    def getWeights(self):  # DriverLib
        return self._weights
    
    @Property
    def inputRanges(self):
        target = zeros((self.inputCount * 2), dtype=npdouble)
        target_p = c_void_p(target.ctypes.data)
        res = self.lib.getinputranges(target_p)
        if res < 0: 
            raise LibraryManagerError(self.lasterror)
        target = target.reshape(self.inputCount, 2)
        return target
    
    @Property
    def outputRanges(self):
        target = zeros((self.outputCount * 2), dtype=npdouble)
        target_p = c_void_p(target.ctypes.data)
        res = self.lib.getoutputranges(target_p)
        if res < 0: 
            raise LibraryManagerError(self.lasterror)
        target = target.reshape(self.outputCount, 2)
        return target
    
    @Property
    def biasbased(self):  # LibManager
        try:
            dim = self.lib.getbiasbased(None)
            target = zeros((dim,), dtype=npint)
            target_p = c_void_p(target.ctypes.data)
            res = self.lib.getbiasbased(target_p)
            if res < 0: 
                raise LibraryManagerError(self.lasterror)
        except AttributeError:
            target = None
        return target
    
    @Property
    def outlinks(self):  # LibManager
        try:
            dim = self.lib.getoutlinks(None)
            target = zeros((dim,), dtype=npint)
            target_p = c_void_p(target.ctypes.data)
            res = self.lib.getoutlinks(target_p)
            if res < 0: 
                raise LibraryManagerError(self.lasterror)
        except AttributeError:
            target = None
        return target
    
    @Property
    def inputNorm(self):
        target = zeros((self.inputCount * 2), dtype=npdouble)
        target_p = c_void_p(target.ctypes.data)
        res = self.lib.getinputnorm(target_p)
        if res < 0: 
            raise LibraryManagerError(self.lasterror)
        
        target = target.reshape(self.inputCount, 2)
        return target
    
    @Property
    def outputNorm(self):
        target = zeros((self.outputCount * 2), dtype=npdouble)
        target_p = c_void_p(target.ctypes.data)
        res = self.lib.getnorm(target_p)
        if res < 0: 
            raise LibraryManagerError(self.lasterror)
        target = target.reshape(self.outputCount, 2)
        return target
    
    def __getstate__(self):
        objdict = self.__dict__.copy()
        objdict["lib"] = None
        objdict["libname"] = ""
        objdict["_lastweights"] = None
        objdict["_callback"] = None
        objdict["targets"] = self.targets
        objdict["propertyName"] = self.propertyName
        objdict["datacount"] = self.lib.dataCount()
        objdict["inputcount"] = self.lib.inputCount()
        objdict["inputs"] = self.fullData
        return objdict
    
    def __setstate__(self, state):
        targets = state.pop("targets", None)
        propertyName = state.pop("propertyName", None)
        datas = state.pop("inputs", None)
        datacount = state.pop("datacount", 0)
        inputcount = state.pop("inputcount", 0)
        self.__dict__.update(state)
        self._setlib(self.libfilename)
        if datas is not None:
            source_p = c_void_p(datas.ctypes.data)
            self.lib.loaddatabase(source_p, datacount, inputcount)      

        if targets is not None:
            self.targets = targets
        if propertyName is not None:
            self.propertyName = propertyName
        
    def Jacobian2long(self, weights=None):    # DriverLib  #, inputs=None, model=None
        """Mise en forme pour les routines d'optimisation.
        Calcule et retourne les sorties et le jacobien du modèle.
        Si le modele est un multimodel parallel, inputs est None, et model est le multimodel.
        Sinon, inputs est la matrice des entrées proposées sur chaque ligne, et model est le modele.
        """
        if weights is None:
            weights = self.weights
        return self.getjacobian(weights, style=2)
    
    def setIdentifiers(self, idlist, datacount=-1): # DriverLib
        res = 0
        if not hasattr(self.lib, "setidentifier"):
            return -1
        if not len(idlist) and datacount > 0:
            idlist = ["data_{0}".format(ind) for ind in range(datacount)]
        for index, val in enumerate(idlist):
            if PY3:
                try:
                    value = val.encode("utf-8")
                except AttributeError:
                    value = str(val).encode("utf-8")
            else:
                value = str(val)
            rtyp = c_char_p(value)
            resloc = self.lib.setidentifier(index, rtyp)
            res |= resloc
        return res
#         if res: 
#             raise LibraryManagerError("getNames error : %s"% res)

    def loadTrainingDataFromArray(self, source, baselen=-1, linelen=-1): # DriverLib
#         def extendList(a, b):
#             a.extend(b)
#             return a
        
        if isinstance(source, ndarray):
            ss = asarray(source.flatten(), npdouble)
            if baselen < 0:
                baselen = source.shape[0]
            if linelen < 0:
                linelen = source.shap[1]
        else:
            if isinstance(source, list) and isinstance(source[0], list):
                if baselen < 0:
                    baselen = len(source)
                if linelen < 0:
                    linelen = len(source[0])
                source = concatenateList(source)
                #reduce(extendList, source)
                
#                 for ind, vlist in enumerate(source):
#                     if not ind:
#                         lst = vlist.copy()
#                     else:
#                         lst.extend(vlist)
#                 source = lst
            try:
                ss = asarray(source, npdouble)
            except:
                ss = asarray(asarray(source).flatten(), npdouble)
#         if isinstance(source, list):
#             if len(source) == 2:
#                 fullsource = source[0].append(source[1])
#                 source = fullsource
#             base = len(source[0])
#             size  = len(source[0][0])
#         else:
#             base, size = source.shape
#         if baselen < 0:
#             baselen = base
#         if linelen < 0:
#             linelen = size
        source_p = c_void_p(ss.ctypes.data)
        return self.lib.loaddatabase(source_p, baselen, linelen)
    
    def loadTrainingDataFromFrame(self, dataFrame, inputCount=-1, outputCount=-1):
        noindex = False
        idname = dataFrame.index.name
        if idname is None:
            noindex = True
            idname = 'index'
        cols = [val.lower() for val in dataFrame.columns]
        coltypes = dataFrame.dtypes
        datacount, colcount = dataFrame.shape
        decal = 0
        lin = 0
        lout = 0
        smilesindex = -1
        if 'smiles' in cols:
            smilesfields = dataFrame.fieldsOf('smiles') 
            decal = len(smilesfields)
            smilesindex = cols.index('smiles')
        if 'inputs' in cols or 'outputs' in cols:
            if 'inputs' in cols:
                infields = dataFrame.fieldsOf('inputs')
                inindexes = dataFrame.indexesOf('inputs')
                lin = len(infields)
            if 'outputs' in cols:
                outfields = dataFrame.fieldsOf('outputs')
                outindexes = dataFrame.indexesOf('outputs')
                lout = len(outfields)
            datas = dataFrame._frame.iloc[:, inindexes+outindexes]
            source = asarray(datas, npdouble).flatten()
            #source = asarray(asarray(dataFrame._frame).iloc[:, inindexes+outindexes].flatten(), npdouble)
            pptyname = outfields[0]
            linelen = lin+lout
        elif outputCount >= 0 and inputCount >= 0 and colcount > inputCount+outputCount:
            pptyname = cols[-1]
            datas = dataFrame._frame.iloc[:, :inputCount + outputCount]
            source = asarray(datas, npdouble).flatten()
            linelen = datas.shape[1]  #inputCount
        else:
            pptyname = cols[-1]
            usecol = [ind for ind, _ in enumerate(dataFrame.columns) if ind != smilesindex]
            datas = dataFrame._frame.iloc[:,usecol]
            source = asarray(datas, npdouble).flatten()
            linelen = datas.shape[1]
        source_p = c_void_p(source.ctypes.data)
        self.lib.loaddatabase(source_p, datacount, linelen)
        # ajouter le chargement des identificateurs 
        self.propertyName = pptyname
        if not noindex:
            self.setIdentifiers(list(dataFrame.index), datacount)
        return datacount, colcount - decal
                 
        
    def loadTrainingData(self, options=None, filename="", filetype="", datarange="", datafields=[], grouplist=[]):
        if (options is None) and not filename: return
        if options is not None:
            filename = options.source
            filetype = options.filetype
            datarange = options.datarange
            datafields = options.datafields
            grouplist = options.grouplist
        datafmt = []
        #for datades in ['inputs', 'outputs']:
        for datades in grouplist:
            datafmt.extend(datafields[grouplist.index(datades)])
        datasource = DataSource(filename, filetype=filetype, datarange=datarange,
            datafmt=datafmt, titles=True)
        values = []
        smiles = []
        ids = []
        datacount = 0
        lid = len(datafields[grouplist.index('identifiers')]) if 'identifiers' in grouplist else 0
        lsmi = len(datafields[grouplist.index('smiles')]) if 'smiles' in grouplist else 0
        lin = len(datafields[grouplist.index('inputs')]) if 'inputs' in grouplist else 0
        lout = len(datafields[grouplist.index('outputs')]) if 'outputs' in grouplist else 0
        for ind, row in enumerate(datasource.source.iterData(datarange=datarange, datafmt=datafmt)):
            if not ind and not isNum(row[-1]):
                continue
            ids.extend(row[:lid])
            smiles.extend(row[lid:lid+lsmi])
            values.extend(row[lid+lsmi:])
            datacount += 1
        datasize = lin + lout
        values = [floatEx(val) for val in values]
        source = asarray(values, npdouble)
        source_p = c_void_p(source.ctypes.data)
        self.lib.loaddatabase(source_p, datacount, datasize)
        if options is not None:
            options.datacount = datacount
            options.datasize = datasize
        # ajouter le chargement des identificateurs 
        self.setIdentifiers(ids, datacount)
        pass        
        
    def loadParameters(self, source):  # DriverLib
        if source is None: return
        if isinstance(source, dict):
            return self.loadFromDict(source)
        elif isinstance(source, str):
            #chargement binaire
            res = False
            with load(source) as sourceopen:
                res = self.loadFromDict(sourceopen)
            if not res:  #self.loadFromDict(load(source)):  #loadWeightsBinary(source):
                # chargement ASCII    
                if not self.loadWeightsASCII(source, self):
                    # chargement XML
                    self.loadWeightsXML(source, self)
        else:
            self.weights = source
#         elif isinstance(source, (list, ndarray)):
#             source = source.copy()
#             if self.weights is None:
#                 self.weights = asarray(source)
#             # chargement d'un vecteur
#             else:
#                 self.weights = source
                    
    def setcallback(self, cbFunc):  # DriverLib 
        """setcallback(cbFunc)
        Installation and uninstallation of the callback function.
        params:
            cbFunc -> Callback function to install. 
                C parameters: (double, long)
                return long
            For uninstallation, this parameter will be "None" 
        """
        self._callback = cbFunc
        if cbFunc:
            curcallback = proxy(cbFunc)
        # definition du type de fonction de callback, conforme a l'attente de la DLL
        # les fonctions de callback sont toujours de type "cdecl"
            self._callbacktype = CFUNCTYPE (c_int, c_double, c_int)#
        # L'enregistrement de la reference "_callback" permet de garder la
        # fonction de callback vivante malgre le garbage collector.
            self._callbacklib = self._callbacktype(curcallback)
        else:
            self._callbacklib = None
        return self.lib.setcallback(self._callbacklib)
    
# #     
# #     @property
# #     def SVDCount(self):
# #         try:
# #             return self.lib.getsvdcount()
# #         except Exception:
# #             return 0
# #     

    @Property
    def trainingResults(self):
        try:
            return self._trainingResults
        except:
            return None
        
#     @property
#     def bestList(self):
#         return self._bestList
    
#     @Property
#     def baseLen(self): # DriverLib
#         try:
#             return self.lib.baselen()
#         except:
#             return 0
#     
#     @Property
#     def dataCount(self): # DriverLib
#         try:
#             return self.lib.dataCount()
#         except:
#             return 0
#     modelCount = dataCount
#     
    @Property
    def modelCount(self):
        return 1
    
    @Property
    def paramNames(self, index): # DriverLib
        return self.getNames(2, index)
    @paramNames.lengetter
    def paramNames(self): # DriverLib
        return self.paramCount
    
    @Property
    def modelNames(self, index): # DriverLib
        return self.getNames(0, index)
    @modelNames.lengetter
    def modelNames(self): # DriverLib
        return self.dataCount
    
    @Property
    def smiles(self, index): # DriverLib
        return self.getNames(1, index) 
    @smiles.lengetter
    def smiles(self): # DriverLib
        return self.modelCount
    
    @Property
    def trainingData(self, index):  # DriverLib
        return self.traindatas()[index]
    @trainingData.lengetter
    def ltrainingData(self):  # Driver
        return self.dataCount
    
    def getTrainingData(self):
        return self.traindatas()
        
    def traindatas(self): # DriverLib
        result = self.fullData
        result.shape = (self.lib.baselen(), self.lib.inputCount())
        return result
    
    @Property
    def fullData(self):
        length = self.lib.getdata(-1, None)
        target = zeros((length,))
        target_p = c_void_p(target.ctypes.data)
        self.lib.getdata(-1, target_p)
        return target
    
#     @Property
#     def trainingData(self, index):
#         length = self.lib.getdata(index, None)
#         target = zeros((length,))
#         target_p = c_void_p(target.ctypes.data)
#         self.lib.getdata(index, target_p)
#         return target
#      
#     @trainingData.lengetter
#     def trainingData(self):
#         return self.lib.baselen()
    
#     @Property
#     def inputs(self, index):
#         length = self.lib.inputCount()
#         target = zeros((length,))
#         target_p = c_void_p(target.ctypes.data)
#         self.lib.getdata(index, target_p)
#         return target
#     
#     @inputs.lengetter
#     def inputs(self):
#         return self.lib.inputCount()
#     
    def gettargets(self):#pour compatibilite
        return self.targets
    
    def settargets(self, values, nonorm=0): # DriverLib
        try:
            source = asarray(values, npdouble)
            source_p = c_void_p(source.ctypes.data)
            res = self.lib.settargets(source_p, int(nonorm))
        except:
            res = 0
        if res < 0: 
            raise DriverLibError(self.lasterror)
    
    @Property
    def targetsmem(self): # DriverLib
        try:
            dim = self.lib.gettargets(0, None)
            target = zeros((dim,), dtype=npdouble)
            target_p = c_void_p(target.ctypes.data)
            res = self.lib.gettargets(-2, target_p)
            if res < 0: 
                raise DriverLibError(self.lasterror)
        except:
            target = None
        return target
            
    @Property
    def targets(self): # DriverLib
        try:
            dim = self.lib.gettargets(-1, None)
            target = zeros((dim,), dtype=npdouble)
            target_p = c_void_p(target.ctypes.data)
            res = self.lib.gettargets(-1, target_p)
            if res < 0: 
                raise DriverLibError(self.lasterror)
        except:
            target = None
        return target
            
    @targets.setter
    def stargets(self, value): # DriverLib
        self._targetfile = ""
        source = None
        if isinstance(value, str) and os.path.exists(value):
            self._targetfile = value
            with open(value, "r") as ff:
                lines = ff.readlines()
                lst0 = [val for val in lines]
                val0 = lst0[0]
                lst = [floatEx(val) for val in lst0]
                if isnan(lst[0]):
                    lst = lst[1:]
                    pptyname = val0
                else:
                    pptyname = os.path.basename(value)
                source = asarray(lst, npdouble)
                self.propertyName = pptyname 
        else:  
            try:
                source = asarray(value, npdouble)
            except: 
                res = -1
#              ValueError:
#             source = None
#             if os.path.exists(value):
#                 self._targetfile = value
#                 with open(value, "r") as ff:
#                     lines = ff.readlines()
#                     lst = [floatEx(val) for val in lines]
#                     source = asarray(lst, npdouble)
#                     self.propertyName = os.path.basename(value)
        if source is not None:
            source_p = c_void_p(source.ctypes.data)
            res = self.lib.settargets(source_p, 0)
        if res < 0: 
            raise DriverLibError(self.lasterror)
        
    def getlastresiduals(self, curw=None, outputs=None):  # DriverLib
        outputs = outputs if outputs is not None else self._outputs
        val = self.targets - outputs
        return val
    
    def getresiduals(self, params=None, style=0): # DriverLib
        if params is not None:
            params = asarray(params)
        else:
            if self.weights is not None:
                params = self.weights
        if params is not None:
            self._lastweights = params.copy
        else:
            raise DriverLibError("params sould not be null in getresiduals")
        try:
            dim = self.lib.getresiduals(None, 0, None)
            source = asarray(params, npdouble)
            source_p = c_void_p(source.ctypes.data)
            target = zeros((dim,), dtype=npdouble)
            target_p = c_void_p(target.ctypes.data)
            res = self.lib.getresiduals(source_p, style, target_p)
            if res < 0: 
                raise DriverLibError(self.lasterror)
        except:
            target = None
        return -target
    
    def freedom(self, index=0):  # DriverLib
        return self.lib.getfreedomindex(index)
    
    def transferICindex(self, params=None, index=-1, trainingStdDev=None, disp=None, acc=0.95):  # DriverLib
        out, leverage = self.transferleverageindex(params, index, False, disp)
        stddev = trainingStdDev if trainingStdDev is not None else self.trainingStddev
        df = self.freedom(index)
        #inter = interval(acc, df)
        if leverage >= 0:
            coeff = stddev * math.sqrt(leverage)
        else:
            coeff = 0
        IC =  interval(acc, df)[1] * coeff
        return out, IC
    
    def transferleverageindex(self, params=None, index=-1, removeindex=False, disp=None):
        if disp is None:
            disp = self._dispersionMatrix
        out, gradient = self.transfergradientindex(params, index=index)
        lev = npdot(npdot(gradient.T, disp), gradient)
        return out, lev
    
    def transferindex(self, params=None, index=-1):
        if isinstance(params, text_type) and (params == 'extra'):
            return [self.transferindex(param, index) for param in self._extraweights]
        if params is not None:
            params = asarray(params, npdouble)
        else:
            if self.weights is not None:
                params = self.weights
        if params is None:
            if self.weights is not None:
                params = self.weights
        if params is not None:
            self._lastweights = params.copy
            params = asarray(params, npdouble)
        else:
            raise DriverLibError("params sould not be null in transferindex")
         
        out = None
        try:
            source = params.copy()
            source_p = c_void_p(source.ctypes.data)
            out = zeros((1,), dtype=npdouble)
            out_p = c_void_p(out.ctypes.data)
            res = self.lib.transferindex(index, source_p, out_p)
            if res < 0: 
                raise DriverLibError(self.lasterror)
             
        except: pass
        if self.outputCount ==1:
            out = out[0]
        return out
 
    def transfergradientindex(self, params=None, index=-1):
        if params is not None:
            params = asarray(params, npdouble)
        else:
            if self.weights is not None:
                params = self.weights
        if params is None:
            if self.weights is not None:
                params = self.weights
        if params is not None:
            self._lastweights = params.copy
            params = asarray(params, npdouble)
        else:
            raise DriverLibError("params sould not be null in transfergradientindex")
         
        out = None
        #target = None
        try:
            paramCount = self.lib.paramCount()
            source = params.copy()  #asarray(params, npdouble)
            source_p = c_void_p(source.ctypes.data)
            gradient = zeros((paramCount,), dtype=npdouble)
            gradient_p = c_void_p(gradient.ctypes.data)
            out = zeros((1,), dtype=npdouble)
            out_p = c_void_p(out.ctypes.data)
            res = self.lib.transfergradientindex(index, source_p, out_p, gradient_p)
            if res < 0: 
                raise DriverLibError(self.lasterror)
             
        except: pass
        if self.outputCount ==1:
            out = out[0]
        return out, gradient
    
#     def transfer(self, inputs=None, params=None): # DriverLib
#         # inputs must be an np.array, type np.double
#         if isinstance(params, str) and (params == 'extra'):
#             return [self.transfer(inputs, param) for param in self.extraweights]
#         if params is not None:
#             params = asarray(params, npdouble)
#         else:  # not isinstance(params, str):
#                 params = self.weights
#             if self.weights is not None:
    
    def transfer(self, inputs=None, style=C.TFR_STD, params=None, **kwds):
        # inputs must be an np.array, type np.double
        indexmove = kwds.get('indexmove', -1)
        indexlist = kwds.get('indexlist', None)
        if (inputs is not None) and (indexmove >= 0) and (indexlist is not None):
            res = [self.transfer(insertInList(inputs, value, indexmove), style, params) for value in indexlist]
#             res = [0] * len(indexlist)
#             for ind, value in enumerate(indexlist):
#                 #inputs[indexmove] = value
#                 res[ind] = self.transfer(insertInList(inputs, value, indexmove), style, params)
            
            return res
        
        if style == C.TFR_GRADIENT:
            return self.transfergradient(inputs, params, **kwds)
        if style == C.TFR_LEVERAGE:
            disp = kwds.get('dispersion', None)
            return self.transferleverage(inputs, params, disp, **kwds)
        if style != C.TFR_STD:
            raise NotImplementedError
        index = kwds.get('index', -1)
        if index >= 0:
            return self.transferindex(params, index)
        
        if isinstance(params, str) and (params == 'extra'):
            return [self.transfergradient(inputs, param) for param in self.extraweights]
        if params is None:
            if self.weights is not None:
                params = self.weights
        if params is not None:
            self._lastweights = params.copy
            params = asarray(params, npdouble)
        if inputs is not None:
            inputs = asarray(inputs)
        try:
            param_p = None if params is None else c_void_p(params.ctypes.data)  #.copy()
#             if params is None:
#                 param_p = None
#             else:
#                 param = params.copy()
#                 param_p = c_void_p(param.ctypes.data)
            inp_p = c_void_p(inputs.ctypes.data)
            out = zeros((self.outputCount,), dtype=npdouble)
            out_p = c_void_p(out.ctypes.data)
            res = self.lib.transfer(inp_p, param_p, out_p)
            if not res:
                if self.outputCount ==1:
                    out = out[0]
            else:
                out = None
        except Exception as err:
            out = None
        return out
    
    def transferEx(self, inputs=None, withCI=0, modelindex=-1):
#         if (self.weights is not None) and (len(self.weights) > self.paramCount):
#             self.mainModel.weights = self.weights
        outs = []
        levs = []
        stds = []
        CIp = []
        CIm = []
        if (self._extraweights is None) or (self._extradispersions is None):
#             out, lev = self.transferLeverages(inputs=inputs)  #original=original
#             out, grad = self.transfergradient(inputs=inputs)
            out, lev = self.transferleverage(inputs=inputs)  #original=original
            outs.append(squeeze(out))
            levs.append(squeeze(lev))
        else:
            if self._extrastddev is None:
                extrastddev = [0 for _ in range(len(self._extraweights))]
            else:
                extrastddev = self._extrastddev
            if 0 <= modelindex < self.extraCount:
                try:
                    weight = self._extraweights[modelindex] 
                    disp = self._extradispersions[modelindex]
                    std = extrastddev[modelindex]
                    out, lev = self.transferLeverages(inputs, weight, disp) 
                    outs.append(out.squeeze())
                    levs.append(lev.squeeze())
                    stds.append(std.squeeze())
                except AttributeError:
                    raise
            else:     
                try:
                    for weight, disp, std in zip(self._extraweights, self._extradispersions, extrastddev):
                        out, lev = self.transferLeverages(inputs, weight, disp)  
                        outs.append(out.squeeze())
                        levs.append(lev.squeeze())
                        try:
                            stds.append(std.squeeze())
                        except AttributeError:
                            stds.append(std)           
                except AttributeError as err:
                    raise
        if not withCI:
            return outs, levs
        students = [self._studentvariable for _ in range(len(outs))] 
        if withCI == 1:
            return outs, levs, stds, students
        for out, lev, std in zip(outs, levs, stds):
            ic = sqrt(lev)*self._studentvariable*std
            CIp.append(out + ic)
            CIm.append(out - ic)
        return outs, CIp, CIm, levs
    
    def transfergradient(self, inputs=None, params=None, **kwds):
        if isinstance(params, str) and (params == 'extra'):
            return [self.transfergradient(inputs, param) for param in self.extraweights]
        if params is not None:
        # inputs must be an np.array, type np.double
            params = asarray(params, npdouble)
        elif self.weights is not None:
                params = self.weights
#         if params is not None:
#             self._lastweights = params.copy
        try:
            param = None if params is None else params.copy()
            self._lastweights = param
            param_p = None if param is None else c_void_p(param.ctypes.data)
            inp_p = c_void_p(inputs.copy().ctypes.data)
            out = zeros((self.outputCount,), dtype=npdouble)
            out_p = c_void_p(out.ctypes.data)
            grad = zeros((self.paramCount,), dtype=npdouble)
            grad_p = c_void_p(grad.ctypes.data)
            res = self.lib.transfergradient(inp_p, param_p, out_p, grad_p)
            if self.outputCount ==1:
                out = out[0]
        except:
            out = grad = None
        return out, grad
    
    def transferleverage(self, inputs=None, params=None, disp=None, **kwds):
        if isinstance(params, text_type) and (params == 'extra'):  #
            return [self.transferleverage(inputs, param, disper) for param, disper in zip(self.extraweights, self.extradispersions)]
        if params is not None:
            params = asarray(params, npdouble)
        else:
            if self.weights is not None:
                params = self.weights
        if params is not None:
            self._lastweights = params.copy
#         else:
#             raise DriverLibError("params sould not be null in transfergradientindex")
        lev = NaN
        try:
            param = None if params is None else params.copy()
            param_p = None if param is None else c_void_p(param.ctypes.data)
            inputs = inputs.copy()
            inp_p = c_void_p(inputs.ctypes.data)
            disper = None if disp is None else disp.copy()
            disper_p = None if disp is None else c_void_p(disper.ctypes.data)
            out = zeros((self.outputCount,), dtype=npdouble)
            out_p = c_void_p(out.ctypes.data)
            lev_p = c_double(lev)
            res = self.lib.transferleverage(inp_p, param_p, disper_p, out_p, byref(lev_p))
            lev = lev_p.value
            if self.outputCount ==1:
                out = out[0]
        except Exception as err: 
            out = lev = NaN
        return out, lev
            
    def getjacobian(self, params=None, style=0): # DriverLib
        if params is not None:
            params = asarray(params)
        else:
            if self.weights is not None:
                params = self.weights
        if params is not None:
            self._lastweights = params.copy
        else:
            raise DriverLibError("params sould not be null in getjacobian")
        #dim = self.lib.get residuals(None, 0, None, None)
        try:
            modelCount = self.lib.dataCount()
            paramCount = self.lib.paramCount()
            source = asarray(params, npdouble)
            source_p = c_void_p(source.ctypes.data)
            target = zeros((modelCount*paramCount,), dtype=npdouble)
            target_p = c_void_p(target.ctypes.data)
            out = zeros((modelCount,), dtype=npdouble)
            out_p = c_void_p(out.ctypes.data)
            
            res = self.lib.getjacobian(source_p, style, target_p, out_p) 
            
            if res < 0: 
                raise DriverLibError(self.lasterror)
            target2 = reshape(target, (modelCount, paramCount))
            target2 = target2[:self.baseLen]
#             ll = self.get dirlen()
#             if 0 < ll < self.modelCount:
#                 target2 = target2[:ll]
        except:
            out = None
            target2 = None
        if style:
            return out, target2
        return target2

    def getTrainType(self):
        return self.lib.gettraintype()
    
    def getStdDev(self, params=None, trainstyle=0, **kwds):
        return self.getCost(params, True, True, trainstyle)
    
    def getCost(self, params=None, donorm=False, root=False, trainstyle=-1, **kwds): # DriverLib  
        """Cost computation.
        params -> model parameters to be used. If None, current weights are used
        donorm -> result normalisation with the data count
        root -> extract the resjult square root (return stdDev)
        trainstyle -> style of cost computation. -1 keep the current trainstyle.
            see CS_xxx constants in monalconst module.
        """
        iparams = None
        if params is not None:
            iparams = asarray(params, npdouble)
        else:
            if self.weights is not None:
                iparams = asarray(self.weights, npdouble)
        if iparams is not None:
            self._lastweights = iparams.copy
        else:
            raise DriverLibError("params sould not be null in getCost")
        try:
            if trainstyle >= 0:
                trainstylemem = self.setTrainStyle(trainstyle)
            dim = self.lib.getcost(None, None)
            source_p = c_void_p(iparams.ctypes.data)
            target = zeros((dim,), dtype=npdouble)
            target_p = c_void_p(target.ctypes.data)
            res = self.lib.getcost(source_p, target_p)
            if res < 0: 
                raise DriverLibError(self.lasterror)
            if donorm:
                target /= self.lib.dataCount()
            if root:
                target = sqrt(target);
        except Exception as err:
            target = None
        finally:
            if trainstyle >= 0:
                self.setTrainStyle(trainstylemem)
        if dim == 1:
            target = target[0]
        return target
        
    def setTrainStyle(self, style):
        return self.lib.settrainstyle(c_int(style))
        
    def getTrainStyle(self):
        return self.lib.gettrainstyle()
    
    def getPRESS(self, params=None, donorm=False, root=False, full=False, 
                 trainstyle=0):
        iparams = None
        if params is not None:
            iparams = asarray(params)
        else:
            if self.weights is not None:
                iparams = self.weights
        if iparams is not None:
            self._lastweights = iparams.copy
        else:
            raise DriverLibError("params sould not be null in getPRESS")
        try:
            svdlen = 0
            if trainstyle >= 0:
                stylemem = self.setTrainStyle(trainstyle)
            else:
                trainstyle = self.getTrainStyle()
            #trainstyle = 0 #C.CS_MODERATE #| C.CS_LIMITED_LEVERAGE
            deltash = 0.0
            if full:
                ZMat = zeros((self.paramCount*self.paramCount,), dtype=npdouble)
                ZMat_p = c_void_p(ZMat.ctypes.data) 
                leverages = zeros((self.lib.baselen(),), dtype=npdouble)
                leverages_p = c_void_p(leverages.ctypes.data)
                residuals = zeros((self.lib.baselen(),), dtype=npdouble)
                residuals_p = c_void_p(residuals.ctypes.data)
                singular = zeros((self.paramCount,), dtype=npdouble)
                singular_p = c_void_p(singular.ctypes.data)
            else:
                ZMat_p = None
                leverages_p = None
                residuals_p = None
                singular_p = None
            svdlen_p = c_long(svdlen)
            deltash_p = c_double(deltash)
            dim = self.lib.outputCount()
            source = iparams  #asarray(iparams, npdouble)
            source_p = c_void_p(source.ctypes.data)
            press = zeros((dim,), dtype=npdouble)
            press_p = c_void_p(press.ctypes.data)
            
            if CC.__version__ >= 21:
                res = self.lib.getpress(source_p, press_p, residuals_p,
                singular_p, leverages_p, ZMat_p, byref(deltash_p), 
                c_int(trainstyle), byref(svdlen_p))  # getpress
            else:
                res = self.lib.getpress(source_p, press_p, residuals_p,
                singular_p, leverages_p, ZMat_p, byref(svdlen_p))  # getpress
            if res < 0: 
                raise DriverLibError(self.lasterror)
            if donorm:
                press /= self.lib.baselen()
            if root:
                press = sqrt(press)
            residuals = -residuals
        except Exception as err:
            print(err)
            press = None
        finally:
            if trainstyle >= 0:
                self.setTrainStyle(stylemem)
        if dim == 1:
            press = press[0]
        if full:
            disp = ZMat.reshape(self.paramCount, self.paramCount)
            return press, residuals, singular, leverages, disp, svdlen_p.value, deltash_p.value
                
        return press

    def getprime(self, params=None, style=0):  # DriverLib
        if params is not None:
            params = asarray(params)
        else:
            if self.weights is not None:
                params = self.weights
        if params is not None:
            self._lastweights = params.copy
        else:
            raise DriverLibError("params sould not be null in getprime")
        dim = self.lib.prime(None, style, None)
        source = asarray(params, npdouble)
        source_p = c_void_p(source.ctypes.data)
        target = zeros((dim,), dtype=npdouble)
        target_p = c_void_p(target.ctypes.data)
        res = self.lib.prime(source_p, style, target_p)
        if res < 0: 
            raise DriverLibError(self.lasterror)
        return target
                    
    @Property
    def director(self):  # DriverLib
        dim = self.lib.getdirector(0)  #None
        target = zeros((dim,), dtype=npint)
        target_p = c_void_p(target.ctypes.data)
        res = self.lib.getdirector(target_p)
        if res < 0: 
            raise DriverLibError(self.lasterror)
        return target
    @director.setter
    def director(self, values):  # DriverLib
        values = list(values)
        if len(values) < self.modelCount:
            values.append(-1)  # modif 14/12/17 'values =' supp
        source = asarray(values, npint)
        source_p = c_void_p(source.ctypes.data)
        res = self.lib.setdirector(source_p)
        if res < 0: 
            raise DriverLibError(self.lasterror)
    
    def initWeights(self, stddev=0.3, bias0=True, seed=None, **kwds):
        if seed is not None:
            self.setRandSeed(seed)
        weights = self.newWeights(stddev=stddev, bias0=bias0)
        self.weights = weights
    
    def newWeights(self, stddev=0.3, bias0=True, doinit=True):  # DriverLibMulti
        if not doinit:
            return self.weights
        if bias0:
            lst = self.biasbased
        else:
            lst = []
        return asarray(self.getNewWeights(stddev, lst))
    
    def settrainingset(self, style=0, index=0, residuals=None):  # DriverLib
        if (residuals is None) or (style == 2):
            source_p = None
        else:
            source = asarray(residuals, float)
            source_p = c_void_p(source.ctypes.data)
        res = self.lib.settrainingset(style, index, source_p)
        if res < 0: 
            raise DriverLibError(self.lasterror)
        return res
                
    def multitrain(self, # DriverLib
            initcount=1, 
            epochs=100, 
            initweights=None, 
            weightsstddev=0.3, 
            seed=None, 
            style=C.TR_S_STD, # 0
            substyle=0, 
            writer=sys.stdout,
            onEndIter=None, 
            onReturnResult=None, 
            trainstyle=0, 
            maxworkers=-2, 
            parallel=True,
            callback=callbackmin,
            **kwds): 
        
        parallel = True  # !!! ATTENTION  mettre a True!!!
        """multitrain(initcount=1, epochs=0, algo="LM_INTEGRATED", initweights=None,
                   seed=-1, callback=None, criterion=C.TCR_COST)  debugfile=None, 
        
        Lancement de plusieurs séquences d'apprentissage
        """
        max_workers = maxWorkers(maxworkers)
        do_DEBUG_TRAIN = (trainstyle & C.CS_DEBUG)  
        if do_DEBUG_TRAIN:
            print("startweights", initweights)
        do_parallel = (parallel and 
                    USE_PARALLEL_MAIN and 
                    (initcount > 1) and 
                    (max_workers > 0) and 
                    not do_DEBUG_TRAIN)
        if not do_parallel:
            # ProcessPoolExecutor ne sait pas gerer ce callback
            self.setcallback(callback)
        self.leverages = None
        self.stoptrain = 0 
        self.cumulEpoch = 0
        if epochs <= 0:
            epochs = 100*(1 + self.paramCount)
        self.setRandSeed(seed)
        if writer is not None and self.isLinear:
            writer.write("Current model is linear.\n")
        if initweights is not None:
            self.loadParameters(initweights)
        else:
            self.weights = self.newWeights(weightsstddev, 
                doinit=(self.weights is None) or (style & C.TR_S_INIT_PARAM))        
        self.memWeights = self.weights.copy()
            
        #self._trainingResults = zeros((initcount, C.TCR_COUNT))
        self._dispersionMatrix = None
        self._conditionning = None
        self._stdtrain = 0
        muststop = 0
        
        # ici on sépare la boucle d'apprentissage en deux boucles.
        # la boucle 1 procede a l'apprentiisage proprement dit, et met en mémoire les résultats
        # la boucle 2 enregistre les resultats        
        #winit = [self.weights.copy()] + [self.newWeights(weightsstddev) for _ in xrange(initcount-1)]
        stylemem = self.setTrainStyle(trainstyle)
        if no_parallel_seed:
            sel = seedList(seed, initcount)
        else:
            sel = [seed for _ in range(initcount)]
            
        fullresults = []
        if do_parallel:
        # traitement en parallele
            with cf.ProcessPoolExecutor(max_workers=max_workers) as executor:  # multitrainNew
                futures = [executor.submit(trainJob, self, 
                                    weightsstddev, 
                                    epochs, 
                                    True,
                                    ind, 
                                    sed) for ind, sed in enumerate(sel)] 
#                                     seed) for ind in range(initcount)] 
 
                for indloc, future in enumerate(cf.as_completed(futures)):
                    res0, res1, res2 = future.result()
                    ind, cur, locepochs, stddev, startweights, trainend, history, locseed, _ = res0
                    resvec = res2
                    self.stoptrain = (onEndIter and onEndIter(history, trainend, ind, indloc, locepochs))
                    if (cur is not None) and not self.stoptrain:
                        self._lastweights = cur.copy()
                    fullresults.append((indloc, res0, res1, res2))
                    if self.stoptrain:
                        break
                if self.stoptrain:
                    for future2 in futures:
                        future2.cancel()
        else:  # traitement en serie
            for indloc in range(initcount):
                res0, res1, res2 = trainJob(self, 
                        weightsstddev, 
                        epochs, 
                        True, 
                        indloc, 
                        seed)        
                ind, cur, locepochs, stddev, startweights, trainend, history, locseed, _ = res0
                self.stoptrain = (onEndIter and onEndIter(history, trainend, ind, indloc, locepochs))
                if (cur is not None) and not self.stoptrain:
                    self._lastweights = cur.copy()
                fullresults.append((indloc, res0, res1, res2))
                if self.stoptrain:
                    break 
        self.setTrainStyle(stylemem)
        nfail = 0
        for res in fullresults:
            indloc = res[0]
            ind, cur, locepochs, stddev, startweights, trainend, \
                history, locseed, code = res[1]  
            if res[2] is None:
                r2 = (None, None, None, None, None, None, None, None, None)
            else:
                r2 = res[2]
            press, residuals, singular, leverages, disp, svdlen, deltash, \
                outputs, jacobians = r2  #
            resvec = res[3]
            if cur is None:
                nfail += 1
                print ("fail", indloc)
                # que faire ici ???
            else: 
                self._lastweights = cur.copy()
                self._outputs = outputs
                self._jacobians = jacobians
                self._stdtrain = stddev #math.sqrt(stddev/self.modelCount)
                self._dispersionMatrix = disp
                self.actualepochs = locepochs
                self.cumulEpoch += locepochs
                
                if onReturnResult:
                    # dbConnect, tablename, indcur, resultvector, startweights, endweights, leverages, dispersion, history
                    restrain = onReturnResult(ind, resvec, startweights, cur, leverages, disp, history, locseed)
            
        if do_DEBUG_TRAIN:
            print("EndWeights", self.weights)
        return nfail
    
    def kerneltrainintegratedmain(self, weights, maxiter, withhistory=False, index=-1):
        assert weights is not None
        assert maxiter 
        code = 0
        mem = weights.copy()
        try:
            res = self.train_integrated(weights, epochs=maxiter, withhistory=withhistory)[0]
            cur, N, stddev, trainend, hist, code = res  
            if withhistory:
                history = sqrt(hist / self.dataCount)
            else:
                history = None  
        except Exception as exres:
            N = -1
            stddev = 0.0
            cur = None
            print("DriverLib.train", exres)
            if not code:
                code = -1
            return cur, N, stddev, None, None, None, None, None, code
        #self.setdebugfile("")
        resvec = None
        disp = None
        if code:
            cur = None
            stddev = 0.0
        else:
            self.actualepochs = N
            if C.USE_INTEGRATED_LEVERAGE:
                resvec = self.computeTrainingResultsIntegrated(curw=cur) 
            else:
                resvec = self.computeTrainingResultsOld(curw=cur) 
#             resvec = self.computeTrainingResults(curw=cur) 
            disp = self._dispersionMatrix.copy() 
             
        return cur, N, stddev, mem, trainend, resvec, disp, history, code
        
    def core_train2(self, weights, maxiter, withhistory=False, index=-1):
        assert weights is not None
        assert maxiter 
        #resvec = disp = None
        code = 0
        mem = weights.copy()
        try:
            res = self.train_integrated(weights, epochs=maxiter, withhistory=withhistory)[0]
            cur, N, stddev, trainend, hist, code = res  
            if withhistory:
                history = sqrt(hist / self.dataCount)
            else:
                history = None  
        except Exception as exres:
            N = -1
            stddev = 0.0
            cur = None
            print("DriverLib.train", exres)
            if not code:
                code = -1
            return index, cur, N, stddev, None, None, None, code
        if code:
            cur = None
            stddev = 0.0
        return index, cur, N, stddev, mem, trainend, history, code
        
    def core_train(self, weights, maxiter, withhistory=False, index=-1):
        assert weights is not None
        assert maxiter 
        #resvec = disp = None
        code = 0
        memweights = weights.copy()
        try:
            res = self.train_integrated(weights, epochs=maxiter, withhistory=withhistory, computeDispersion=True)
            cur, N, stddev, trainend, hist, code = res[0] 
            if withhistory:
                history = sqrt(hist / self.dataCount)
            else:
                history = None  
        except Exception as exres:
            N = -1
            stddev = 0.0
            cur = None
            print("DriverLib.train", exres)
            if not code:
                code = -1
            return index, cur, N, stddev, None, None, None, code
        if code:
            cur = None
            stddev = 0.0
            
        res0 = (index, cur, N, stddev, memweights, trainend,  history, code)
        #target, residuals, singular, leverages, disp, svdlen, deltash, outputs, jacobians = res[1]  #
        #resvec = res[2] 
        return res0, res[1], res[2]
        
    def kerneltrainintegratedmain3(self, weights, maxiter, withhistory=False, index=-1):
        assert weights is not None
        assert maxiter 
        code = 0
        memweights = weights.copy()
        try:
            res = self.train_integratedex(weights, epochs=maxiter, withhistory=withhistory)
            cur, N, stddev, press, residuals, leverages, disp, jacob, trainend, deltash, svdlen, singular, hist, code = res
            NN = self.dataCount
            qq = self.paramCount
            disp = reshape(disp, (qq, qq))
            jacob = reshape(jacob, (NN, qq))
            #print "dispersion", disp.shape
            #print "jacob", jacob.shape
            if withhistory:
                history = sqrt(hist / self.dataCount)
            else:
                history = None  
        except Exception as exres:
            N = -1
            stddev = 0.0
            cur = None
            print("DriverLib.trainex", exres)
            if not code:
                code = -1
            return index, cur, N, stddev, None, None, None, None, None, None, None, None, None, None, None, code
        if code:
            cur = None
            stddev = 0.0
        return index, cur, N, stddev, press, memweights, residuals, leverages, disp, jacob, trainend, deltash, svdlen, singular, history, code
        
#     def aftertrain2(self, onEndIter, indloc, ind, cur, locepochs, 
#                    stddev, startweights, trainend, history, code):  #resvec, disp, history, code):
#         """Action apres apprentissage, a l'interieur de la boucle d'exploitation des jobs paralleles
#         """
#         muststop = (cur is None) or (onEndIter and onEndIter(history, trainend, ind, locepochs))
#         if not muststop:
#             if cur is None:
#                 outputs = None 
#                 jacobians = None
#             else:
#                 self._lastweights = cur.copy()
#                 outputs, jacobians = self.getjacobian(cur, style=2)
#                 self._stdtrain = math.sqrt(stddev/self.modelCount)
#         return muststop, indloc, ind, cur, locepochs, stddev, startweights, trainend, history, outputs, jacobians, code
        
    def aftertrain(self, indloc, ind, cur, locepochs, 
                   stddev, startweights, trainend, history, code): 
        """Action apres apprentissage, a l'interieur de la boucle d'exploitation des jobs paralleles
        """
        #self.stoptrain = (onEndIter and onEndIter(history, trainend, ind, locepochs))
        muststop = (cur is None) or self.stoptrain
        if not muststop: # and (cur is not None):
            self._lastweights = cur.copy()
        return muststop, indloc, ind, cur, locepochs, stddev, startweights, trainend, history, code 
        
    def train_integrated(self, params, epochs=-1, withhistory=False, 
            computeDispersion=False, trainstyle=-1):
        hist = None
        hist_p = None
        trainEnd = 0
        
        params = asarray(params, npdouble)
        params_p = c_void_p(params.ctypes.data)
        epoch_t = c_long(epochs)
        trainEnd_t = c_long(trainEnd)
        if withhistory:
            hist = zeros((epochs+1,))
            hist.fill(NaN)  # modif du 9/02/15
            hist_p = c_void_p(hist.ctypes.data)
        if trainstyle >= 0:
            trainstylemem = self.setTrainStyle(trainstyle)
        code = self.lib.train(params_p, byref(epoch_t), byref(trainEnd_t), hist_p)
        if trainstyle >= 0:
            self.setTrainStyle(trainstylemem)
        self.setWeights(params)
        if computeDispersion:
            resvec = self.computeTrainingResults(curw=params, actualepochs=epoch_t.value)
            respress = self.getPRESS(params, donorm=True, root=True, full=True)
            residuals = respress[1]
            stddev = sqrt((residuals**2).sum()/self.dataCount)
            outputs, jacobians = self.getjacobian(params, style=2)
            press, residuals, singular, leverages, disp, svdlen, deltash = respress
            resfull = (press, residuals, singular, leverages, disp, svdlen, deltash, outputs, jacobians)
        else:
            resfull = resvec = outputs = jacobians = None
            stddev = self.getStdDev(params)         
        trainEnd = trainEnd_t.value
        return (params, epoch_t.value, stddev, trainEnd, hist, code), resfull, resvec
        
#     def train_integratedOld(self, params, epochs=-1, withhistory=False, 
#             computeDispersion=False):
#         hist = None
#         hist_p = None
#         trainEnd = 0
#         
#         source = asarray(params, npdouble)
#         source_p = c_void_p(source.ctypes.data)
#         epoch_t = c_long(epochs)
#         trainEnd_t = c_long(trainEnd)
#         if withhistory:
#             hist = zeros((epochs+1,))
#             hist.fill(NaN)  # modif du 9/02/15
#             hist_p = c_void_p(hist.ctypes.data)
#         if self.codeVersion >= 28:
#             code = self.lib.train(source_p, byref(epoch_t), byref(trainEnd_t), hist_p)
#         elif self.codeVersion >= 16:
#             code = self.lib.train(source_p, byref(epoch_t), byref(trainEnd_t), hist_p, -1)
#         elif self.codeVersion >= 6:
#             code = self.lib.train(source_p, byref(epoch_t), byref(trainEnd_t), hist_p)
#         elif self.codeVersion >= 5:
#             code = self.lib.train(source_p, byref(epoch_t), hist_p)
#         else:         
#             code = self.lib.train(source_p, byref(epoch_t))
#         self.setWeights(params)
#         if computeDispersion:
#             resvec = self.computeTrainingResults(curw=params, actualepochs=epoch_t.value)
#             respress = self.getPRESS(params, donorm=True, root=True, full=True)
#             residuals = respress[1]
#             stddev = sqrt((residuals**2).sum()/self.dataCount)
#             outputs, jacobians = self.getjacobian(params, style=2)
#             press, residuals, singular, leverages, disp, svdlen, deltash = respress
#             resfull = (press, residuals, singular, leverages, disp, svdlen, deltash, outputs, jacobians)
#         else:
#             resfull = resvec = outputs = jacobians = None
#             stddev = self.getStdDev(source)         
#         trainEnd = trainEnd_t.value
#         return (source, epoch_t.value, stddev, trainEnd, hist, code), resfull, resvec

    def train_integratedex(self, params, epochs=-1, withhistory=False, 
            style=-1):
        hist = None
        trainEnd = 0
        svdlen = 0
        deltash = 0.0
        source = asarray(params, npdouble)
        source_p = c_void_p(source.ctypes.data)
        epoch_t = c_long(epochs)
        trainEnd_t = c_long(trainEnd)
        ZMat = zeros((self.paramCount*self.paramCount,), dtype=npdouble)
        ZMat_p = c_void_p(ZMat.ctypes.data) 
        Jacob = zeros((self.paramCount*self.dataCount,), dtype=npdouble)
        Jacob_p = c_void_p(Jacob.ctypes.data) 
        leverages = zeros((self.lib.baselen(),), dtype=npdouble)
        leverages_p = c_void_p(leverages.ctypes.data)
        residuals = zeros((self.lib.baselen(),), dtype=npdouble)
        residuals_p = c_void_p(residuals.ctypes.data)
        singular = zeros((self.paramCount,), dtype=npdouble)
        singular_p = c_void_p(singular.ctypes.data)
        svdlen_p = c_long(svdlen)
        deltash_p = c_double(deltash)
        dim = self.lib.outputCount()
        target = zeros((dim,), dtype=npdouble)
        target_p = c_void_p(target.ctypes.data)
        cost = zeros((dim,), dtype=npdouble)
        cost_p = c_void_p(target.ctypes.data)
        
        if withhistory:
            hist = zeros((epochs+1,))
            hist.fill(NaN)  # modif du 9/02/15
            hist_p = c_void_p(hist.ctypes.data)
        else:
            hist = None
            hist_p = None
            
        if self.codeVersion < 33:
            code = 0xFFF
        else:
            code = self.lib.trainex(source_p, byref(epoch_t), byref(trainEnd_t),
                hist_p, cost_p, target_p, residuals_p, singular_p, leverages_p, 
                ZMat_p, Jacob_p, byref(deltash_p), byref(svdlen_p))
            if code < 0: 
                raise DriverLibError(self.lasterror)
            self.setWeights(source)
            nlen = self.lib.baselen()
            press = sqrt(target / nlen)
            stddev = sqrt(cost / nlen)

        return source, epoch_t.value, stddev, press, residuals, leverages, \
            ZMat, Jacob, trainEnd_t.value, deltash, svdlen, singular, hist, \
            code

    def train(self, 
              startWeights=None, 
              epochs=0, 
              index=-1, 
              callback=callbackmin, 
              **kwds):                                    # DriverLib
        #algo="LM_INTEGRATED", disp=0, weightsstddev=0.3,
        """Lancement d'un apprentissage.
        Utilise les algorithmes d'optimisation integres.
        callback est une fonction  de callback prototype C: typedef long (*cbk)(real, long);
        """
        if callback == "debug":
            callback = callbackminprint
        trainstyle = kwds.get('trainstyle', 0)
        # setcallback assure la conversion de type de la fonction python 
        # callback vers le type C
        self.setcallback(callback)
        epochs = kwds.get('maxiter', epochs)
#         epochs = epochs if not 'maxiter' in kwds else kwds['maxiter']
        Nt = -1 
#         if 'style' in  kwds:
#             style = kwds["style"]
#         else:
#             style = algo
        #trainend = 0         
        vecPrime = self.getjacobian  # a revoir pour LOO et BS
        residuals = self.getresiduals  # a revoir pour LOO et BS
        fcost = self.getCost
        prime = self.getprime
        
        if startWeights is None:
            startWeights = self.weights
#         startWeights = self.newWeights(weightsstddev, 
#                 doinit=(startWeights is None) or (style & C.TR_S_INIT_PARAM))
        if epochs:
            self._dispersionMatrix = None
            self._conditionning = None
            self._stdtrain = 0
#         if disp:
#             sys.stderr.write("\nInformation: training %s, max iterations %d\n" % ("LM_INTRGRATED", epochs))
        fopt = 0.0
        popt = None
        try:
#             if style == "LM_INTEGRATED":
            res = self.train_integrated(startWeights, epochs=epochs, 
                trainstyle=trainstyle)[0]
            popt, Nt, fopt, _, _, code = res  #trainend
#             elif style == "LM_NT":  # LM by JLP
#                 res = optim_leastsq(func=residuals, x0=startWeights, 
#                   maxiter=maxiter, Dfun=vecPrime, 
#                 full_output=1, disp=disp, call back=call back)#
#                 popt, code, Nt, _, _, fopt = res  #Nf, Ng, 
#             elif style == "BFGS_NT":  # BFGS by JLP
#                 res = optim_bfgs(fcost, startWeights,
#                     maxiter=maxiter, Dfun=prime, 
#                     full_output=1, disp=disp, call back=call back)#
#             elif not OKScipy:
#                 raise Exception("Not implemented yet without Scipy")
#             else:
#                 if style == "LM":  # LM by SciPy
#                     res = optimize.leastsq(func=residuals, x0=startWeights, 
#                          maxfev=maxiter,  Dfun=vecPrime, factor=97, full_output=1)  #
#                     popt, _, infodict, _, _ = res #cov, mesg, ier
#                     Nt = infodict["nfev"]
#                     fopt = npsum(infodict['fvec']**2)/2
#                 elif style == "CG":  # CG by SciPy
#                     res = optimize.fmin_cg(f=fcost, x0=startWeights,
#                         fprime=prime, maxiter=maxiter, 
#                         full_output=1, disp=disp, call back=call back)
#                     popt, fopt, Nt, _, _ = res #Ng, warnflag
#                 elif style == "BFGS":  # BFGS by SciPy
#                     res  = optimize.fmin_bfgs(fcost, startWeights, fprime=prime, 
#                         maxiter=maxiter, full_output=1, disp=disp, call back=call back)
#                     popt, fopt, _, _, Nt, _, _ = res  #gopt, Bopt, Ng, warnflag
#                     popt, code, Nt, _, _, fopt = res #Nf, Ng, 
#                 elif style == "SPX":  # SPX by SciPy
#                     res = optimize.fmin(fcost, startWeights,
#                         maxiter=maxiter, full_output=1, disp=disp, 
#                         call back=cal lback)
#                     popt, fopt, Nt, _, _, _  = res #funcalls, warnflag, allvecs
#             else:
#                 popt = None
        except Exception as exres:
            print("DriverLib.train", exres)
        self._lastweights = popt.copy()
        N = self.modelCount
        self._stdtrain = math.sqrt(fopt/N)
        if code:
            return None, code, 0.0
        return popt, Nt, fopt  

    def reverse_train(self, outputtarget, inputs=None, freeinputs=None, 
        epochs=None, withhistory=False, debug=0, **kwds):
        # kwds est une poubelle
        innames = None
        if isinstance(inputs, dict):
            innames = list(inputs.keys())
            inputs = list(inputs.values())
        inputs = asarray(inputs, npdouble)
        assert inputs.shape[0] == self.inputCount
        inputs_p = c_void_p(inputs.ctypes.data)
        
        if freeinputs is None or freeinputs == 'all':
            freeinputs = ones((self.inputCount,), dtype=nplong)
        elif isinstance(freeinputs, dict):
            newfreeinputs = zeros((self.inputCount,), dtype=nplong)
            for index, name in enumerate(self.inputNames):
                if name in freeinputs:
                    newfreeinputs[index] = freeinputs[name]
            freeinputs = newfreeinputs
        elif isinstance(freeinputs, list) and len(freeinputs):
            newfreeinputs = zeros((self.inputCount,), dtype=nplong)
            for value in freeinputs:
                if isinstance(value, str):
                    index = self.inputNames.index(value)
                elif isinstance(value, int):
                    index = value
                if index >= 0:
                    newfreeinputs[index] = 1
            freeinputs = newfreeinputs
        else:
            freeinputs = asarray(freeinputs, nplong)
            
        freeinputs_p = c_void_p(freeinputs.ctypes.data)
        outputtarget_p = c_double(outputtarget)
        if epochs is None:
            epochs = 100
        epochs_p = c_long(epochs)
        ender = 0
        ender_p = c_long(ender)
        if withhistory:
            history = zeros((epochs+1,), dtype=npdouble)
            for ind in range(history.shape[0]):
                history[ind] = float('nan')
            history_p = c_void_p(history.ctypes.data)
        else:
            history_p = None 
        self.lib.settrainstyle(debug)
        self.lib.reversetrain(inputs_p, freeinputs_p, byref(outputtarget_p), 
            byref(epochs_p), byref(ender_p), history_p)
        epochs = epochs_p.value  
        ender = ender_p.value  
        if innames is not None:
            inputs = {name: val for name, val in zip(innames, inputs)}
        if withhistory:
            return inputs, epochs, ender, history[:epochs+1]
        else:
            return inputs, ender
    
    def hasDebug(self, trainstyle):
        return 1 if (trainstyle & C.CS_DEBUG) else 0
    
    def trainLOO(self, initlist=None, inittype=C.INIT_START_PARAM, 
        callback2=None, onReturnResult=None, selected_LOO=None, 
        initcount=1, weightsstddev=0.1, epochs=0, callback=None, 
        LOOcriterion=None, LOOThreshold=0, trainstyle=-1, maxworkers=-2, 
        parallel=True, seed=None):
        
        return self.trainBS(initlist=initlist, inittype=inittype, 
            weightsstddev=weightsstddev, initcount=initcount, epochs=epochs, 
            BStype=RE_LOO, callback=callback, callback2=callback2, 
            onReturnResult=onReturnResult, selected_LOO=selected_LOO, 
            LOOcriterion=LOOcriterion, LOOThreshold=LOOThreshold, 
            trainstyle=trainstyle, maxworkers=maxworkers, 
            parallel=parallel, seed=seed) 
        
    def trainSuper(self, initlist=None, inittype=C.INIT_START_PARAM, 
        weightsstddev=0.3, initcount=1, BScount=100, epochs=0, 
        callback=None, callback2=None, onReturnResult=None, trainstyle=-1, 
        maxworkers=-2, parallel=True, seed=None): 
        
        return self.trainBS(initlist=initlist, inittype=inittype, 
            weightsstddev=weightsstddev, initcount=initcount, BScount=BScount, 
            epochs=epochs, BStype=RE_SUPER, callback=callback, 
                callback2=callback2, onReturnResult=onReturnResult, 
                trainstyle=trainstyle, maxworkers=maxworkers, parallel=parallel)
        
    def trainBS(self, initlist=None, inittype=C.INIT_START_PARAM, weightsstddev=0.1,  #startParams=None, 
        initcount=1, BScount=100, epochs=0, BStype=RE_NONE, callback=None, 
        callback2=None, onReturnResult=None, selected_LOO=None, 
        LOOcriterion=None, LOOThreshold=0, trainstyle=-1, maxworkers=-2, 
        parallel=True, seed=None): 
        # callback est utilise dans le train de base
        # callback2 est utilise dans les boucles externes
        # selected_LOO sert a limiter le nb de run pour debug
        
        computeDispersion = True 
#        resultMatrix = None
        # computeDispersion est mis a Vrai pour permettre le calcul du LOO_leverage
        max_workers = maxWorkers(maxworkers)
        do_selected_LOO = (selected_LOO is not None) and len(selected_LOO)
        do_parallel = (parallel and 
                    # USE_PARALLEL and 
                    USE_PARALLEL_BS and 
                    not do_selected_LOO and 
                    (max_workers > 0) and 
                    not self.hasDebug(trainstyle))# and (BStype < RE_SUPER) 
        
        if self.weights is None:
            originalParams = None
        else:
            originalParams = self.weights.copy()
        muststop = False
#        self.resultMatrix = None
        settrainindex = 0
        stylemem = 0
        extraweights = None
        extraweightslist = []
        leverageslist = []
        try:
            if BStype == RE_LOO:  # LOO
                BScount = self.dataCount
                settrainindex = 2
            elif BStype in BSGROUP:
                settrainindex = 1 
            elif BStype == RE_SUPER:
                initcount = len(initlist)
                settrainindex = 0
                LOOcriterion = None
            if not USE_SUPER_TRAIN:
                trainstyle = trainstyle & (~ C.CS_SUPER_TRAIN)
            stylemem = self.setTrainStyle(trainstyle)
            extraweights = zeros((BScount, self.paramCount))
            extraweightslist = zeros((initcount, BScount, self.paramCount))
            leverages = zeros((BScount,)) if computeDispersion else None
            leverageslist = zeros((initcount, BScount))
            outs = zeros((BScount,)) if computeDispersion else None
            presses = zeros((BScount, )) if computeDispersion else None
            if (BStype >= RE_LOO) and do_selected_LOO:
                BScount = len(selected_LOO)
            if epochs <= 0:
                epochs = 10*(1 + self.paramCount)
            
            if BStype == RE_SUPER:
                residuals = self.getresiduals(style=C.CS_SUPER_TRAIN).copy()
            elif not self.weights is None:
                residuals = self.getresiduals()  # ici le 15/12/18
            else:
                residuals = None
            targets = self.targets.copy()
            
            loccount = 0
            reslist = []
#             if not onReturnResult:
#                 resultMatrix = zeros((BScount, initcount), dtype=object)
                
            if no_parallel_seed:
                seedlistlist = []
                for _ in range(BScount):
                    seedloc = random.randint(0x10, 0xFFFFFFFF)
                    seedlist = seedList(seedloc, initcount)
                    seedlistlist.append(seedlist)
                
            else:
                seedlistlist = [[seed for _ in range(initcount)] for _ in range(BScount)]
            
            
            if do_parallel:  
                # debut boucle parallele
                with cf.ProcessPoolExecutor(max_workers=max_workers) as executor:                    
                    futures = [executor.submit(trainJobBS, self, initlist, 
                        inittype, initindex, residuals, BStype, settrainindex, 
                        epochs, molindex, computeDispersion, weightsstddev, 
                        targets, locseed)
                        for molindex, seedlist in enumerate(seedlistlist) 
                        for initindex, locseed in enumerate(seedlist)]  
                    cumul0 = 0
                    cumul1 = 0
                    for future in cf.as_completed(futures):
                        res = future.result()
                        reslist.append(res)
#                         if resultMatrix is not None:
#                             resultMatrix[molindex, initindex] = res
                        if callback and (not cumul0 % initcount):
                            cumul1 += 1
                            cont = callback(cumul1)
                            if isinstance(cont, tuple):
                                cont = cont[0]
                            muststop = not cont
                        cumul0 += 1
                        if muststop:
                            break
                    if muststop:
                        for future in futures:
                            future.cancel()

            
            # fin boucle parallele
                if SORT_LOO_BS_PARALEL_TRAIN:
                    reslist.sort(key=lambda x: x[4:6])       
                # tri pour retrouver l'ordre initial    
                for res in reslist:
                    molindex, initindex = res[4:6]
                    if onReturnResult:
                        onReturnResult(*res)
#                 if not muststop and (resultMatrix is not None):
#                     for molindex in range(BScount):
#                         resvec = list(resultMatrix[molindex])
#                         resvec = sorted(resvec, key=lambda x:x[4])
#                         resultMatrix[molindex] = resvec
                     
                
            else:  # traitement en serie
                for molindex, sedlist in enumerate(seedlistlist):  #itercount:
                    for initindex, locseed in enumerate(sedlist):  #range(initcount):
                        res = trainJobBS(self, initlist, inittype, initindex, 
                            residuals, BStype, settrainindex, epochs, molindex, 
                            computeDispersion, weightsstddev, targets, locseed)  #, trainstyle
                        reslist.append(res)
#                         if resultMatrix is not None:
#                             resultMatrix[molindex, initindex] = res[:-1]
                        if onReturnResult:
                            onReturnResult(*res)
                                                
                    if callback:
                        cont = callback(molindex+1)
                        if isinstance(cont, tuple):
                            cont = cont[0]
                        muststop = not cont
                    if muststop:
                        break
                    if muststop:
                        break
                            
#             if not muststop and (resultMatrix is not None):
#                 # Ici on traite les résultats enregistrés dans resultMatrix
#                 for index in range(BScount):
#                 # pour chaque molecule
#                     indloc = index if BStype == 3 else BStype
#                     self.settrainingset()
#                     if initcount <= 1:
#                         loccount = 1
#                         val = resultMatrix[index, 0] 
#                         code, epochs, ini_params, paramMem, molindex, \
#                             initindex, indexReturn, cost, press, out, lev, \
#                             resvec, disp, _, _ = val
#                         if computeDispersion:
#                             leverages[index] = lev
#                             outs[index] = out
#                             presses[index] = press 
#                     else:
#                         self.settrainingset(settrainindex, indloc, residuals)
#                         # on cree la liste verticale des resultats de la molecule definie par index
#                         lst = list(resultMatrix[index])
#                         # on trie si necessaire la liste locresult sur la valeur du premier element
#                         try:
#                             # on trie la liste a trier en fonction du critere 
#                             sortindex = C.criterindex[LOOcriterion]
#                             locresult = sorted(lst, key = lambda x: abs(x[sortindex]))
#                         except:
#                             locresult = lst
#                         self.settrainingset()
#                         for ind1, res in enumerate(locresult):
#                             # renseignement de la liste des poids par ordre de tri
#                             extraweightslist[ind1][index] = res[C.TRR_PARAMS].copy()
#                             leverageslist[ind1][index] = res[C.TRR_LEV]
#                             resultMatrix[index, ind1] = res
#                             
#                         if len(locresult):
#                             code, epochs, ini_params, paramMem, molindex, initindex, _, cost, press, out, lev, resvec, disp, _, _ = locresult[0]  #[1] returnindex, , inittype
#                             if computeDispersion:
#                                 leverages[index] = lev
#                                 outs[index] = out
#                                 presses[index] = press
#                         else:
#                             # cas ou tous les apprentissages ont echoue
#                             code, epochs, paramMem, molindex, initindex = (-1, -1, None, index, -1)
#                         assert(index == molindex)
#                     
#                     if epochs and not code:
#                         extraweights[index] = paramMem.copy()
#                         loccount += 1
#                     if callback2:
#                         cont = callback2(index + 1)
#                         if isinstance(cont, tuple):
#                             cont = cont[0]
                            
#             else:
#                 for res in reslist:
#                     code, epochs, ini_params, params, index, initindex, indexReturn, stddev, press, out, lev, resvec, disp, inittype , target, locseed = res
                
            self.settrainingset()
        except Exception as err:
            print(err)
            raise
        finally:
            self.setTrainStyle(stylemem)
            self.weights = originalParams
            self._extraweights = extraweights
            self._extraweightslist = extraweightslist
            self.leverages = leverageslist #leverages
#            self.outputs = outs
            self.presses = presses
#            self.resultMatrix = resultMatrix
        self.settrainingset()
        return loccount  #, resultMatrix   

def onGetNewWeights(model, weightsstddev):
    return model.newWeights(weightsstddev)
    
def kerneltrainintegratedmain(model, weights, maxiter, withhistory, index=-1):
    res = model.kerneltrainintegratedmain(weights, maxiter, withhistory, index)
    if index < 0:
        return res
    return (index,) + res

def core_train2(model, weights, maxiter, withhistory, index=-1):
    return model.core_train2(weights, maxiter, withhistory, index)

# def core_train(model, weights, maxiter, withhistory, index=-1):
#     return model.core_train(weights, maxiter, withhistory, index)

def kerneltrainintegratedmain3(model, weights, maxiter, withhistory, index=-1):
    return model.kerneltrainintegratedmain3(weights, maxiter, withhistory, index)
#     if index < 0:
#         return res
#     return (index,) + res

# def kerneltraining(model, weights, maxiter, withhistory, index=-1):
#     res = model.kerneltraining(weights, maxiter, withhistory)
#     if index < 0:
#         return res
#     return (index,) + res
def checklib(libfilename):
    model = DriverLib(libfilename)
    print(repr(model))
    
def checkpicklelib(libfilename, filename, filetype, datarange, datafields, grouplist):
    from io import BytesIO
    import pickle as pickle
    
    model = DriverLib(libfilename)
    model.loadTrainingData(None, filename, filetype, datarange, datafields, grouplist)
    print(repr(model))
    print(model.traindatas())
    stream = BytesIO()
    pickle.dump(model, stream)
    stream.flush()
    stream.seek(0)
    model2 = pickle.load(stream)
    stream.close()
    print(repr(model2))
    print(model2.traindatas())
#     return model
    
    #for i in range(model.baselen):

#==============================================================================
if __name__ == "__main__":
    
#     source = "/Users/jeanluc/dockerold/workdir/libbjma244t25_si3_5n.so"
#     checklib(source)
    
#     loadTrainingData(self, options=None, filename="", filetype="", datarange="", datafields):
    source = "/Users/jeanluc/Documents/LiClipseWorkspace/Metaphor_Test/src/metaphor/test_monal/testFiles/libnn_5_5_1.so"
    # "/Users/jeanluc/nn1/workdir/libnn_5_5_1.so"
    filename = "/Users/jeanluc/Documents/LiClipseWorkspace/Metaphor_Test/src/metaphor/test_monal/testFiles/L_153.csv"
    #"/Users/Shared/workspace/Monal_Test/src/testFiles/L_153.xlsx"
    filetype ="csv"
    datarange = "DATA"
    datafields = [[0, 1, 2, 3, 4], [5]]
    grouplist = ['inputs', 'outputs']
    #             (libfilename, filename, filetype, datarange, datafields, grouplist)
    checkpicklelib(source, filename, filetype, datarange, datafields, grouplist)
    
    model = DriverLib(source)

    model.setRandSeed(194999)
    ww = model.newWeights(0.3)
    #residu = model.getresiduals(ww)
    #print "residu", residu
    costini = model.getCost(ww)
    #ww2, epochs, cost, code
    #bests, saveres = model.multitrain(initcount=5, initweights=ww, maxiter=100)
    #bestmem = bests[0][0]
    #print "bests"
    #for val in bests:
    #    print "", val
    
#     from io import StringIO
#     import pickle as pickle
#     single = 1
# #     source = "C:\\Projets\\GM\\TestAD\\Base115new\\base115new_chi0iso00n.dll"
# #     source = "/Users/Shared/GMWorkspace/BaseA109/libbasea109_chi03n.so"
#     source = "/Users/jeanluc/nn1/workdir/libnn_9_5_1.so"
#     datafile = "/Users/Shared/workspace/Monal_Test/src/testFiles/B321_19d_JLPloix.xlsx"
# 
# #codeversion
#     
#     model = DriverLib(source)
#     #model.settrainingset()
#     model.setRandSeed(1947)
#     #model.targets = targetfile
#     
# #     stream = StringIO()
# #     pickle.dump(model, stream)
# #     stream.flush()
# #     stream.seek(0)
# #     #stream2 = StringIO(stream.getvalue())
# #     model2 = pickle.load(stream)
# #     stream.close()
#     #model2 = DriverLib(source)
#     #model.settrainingset()
#     #model.setRandSeed(1947)
#     #model2.targets = targetfile
#     
#     print("property:", model.propertyName)
#     print()
#     print("moduleName", model.moduleName)
#     print("monalVersion", model.monalVersion)
#     print("created", model.created)
#     print("mark", model.mark)
#     print("base", model.base)
#     print("linear", model.isLinear)
#     print("paramlCount", model.paramCount)
#     print("modelCount", model.modelCount)
#     print("inputCount", model.inputCount)
#     print("outputCount", model.outputCount)
#     print("hidden", model.hidden)
#     print("biasBased", model.biasbased)
#     print("outlinks", model.outlinks)
#     #model2.targets = [0 for _ in range(model.modelCount)]
#     #print "target 1", model.targets
#     #print "target 2", model2.targets
#     print("outputnorm", model.outputnorm)
#     print("input norm", model.inputnorm)
# 
# 
# 
# #     print("norm 2", model2.outputnorm)
#     print(model.lib)
# #     print(model2.lib)
#     
#     #if single:
#     model.setRandSeed(194999)
#     ww = model.newWeights(0.3)
#     #residu = model.getresiduals(ww)
#     #print "residu", residu
#     costini = model.getCost(ww)
#     #ww2, epochs, cost, code
#     #bests, saveres = model.multitrain(initcount=5, initweights=ww, maxiter=100)
#     #bestmem = bests[0][0]
#     #print "bests"
#     #for val in bests:
#     #    print "", val
#     
#     
#     
#     
#     cur, epochs, cost, trainend, code = model.kerneltrainintegratedmain(ww, 150)
#     #cost2 = model.getcost(cur)
#     #press_old = model.getPRESS_(cur, None, True, True)
#     press = model.getPRESS(cur, 1, 1)
#     print("cost ini  ", costini)
#     print("cost final", cost)
#     print("PRESS", press)
#     #print "PRESS old", press_old
#     print("epochs", epochs)
#     #else:
#         #os.system("ipcluster start")
#         #lst = multitrainintegrated(model, targetfile, None, 8, 100)
#         #for val in lst:
#         #    print val
#         #for ind, (cur, N, fopt, mem, code) in enumerate(lst):
#         #    print ind, N, fopt
#         
#         #os.system("ipcluster stop")
#     
#     #model = None
#     
    print("done")


#-*- coding: ISO-8859-15 -*-
#===============================================================================
# $Id: _model.py 4819 2018-11-02 05:32:25Z jeanluc $
#  Module  _model
#  Projet MonalPy
# 
#  Implementation python de monal
#
#  Author: Jean-Luc PLOIX  -  NETRAL
#  Decembre 2009
#
# Copyright 2016 Jean-Luc PLOIX
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

from numpy import zeros, identity, array, asarray
from numpy.random import randn
try:
    import cPickle  as pickle
except ImportError:
    import pickle
import os, sys
import json
from datetime import datetime
from ...nntoolbox.utils import floatEx
#from nntoolbox._2and3 import isstr
from six import string_types, print_, text_type
from ..Property import Property, Properties, IndexedProperty, makefct
from ..util.monalbase import Component, MonalError 
from ..util.monaltoolbox import isStrNum
from ..util.utils import safemakedirs
from .. import monalconst as C
from ..util import leverage
from ..driver._optimizer import optim_bfgs #optim_leastsq, 

class ModelBase(Component):
    def __init__(self, owner=None, name=""):
        super(ModelBase, self).__init__(owner, name)
        dt = datetime.now()
        self._modelType = 0

    @Property
    def modelCount(self):
        return 1
    
    @Property
    def modelType(self):
        return self._modelType
    @modelType.setter
    def modelType(self, value):
        self._modelType = value
        
    @Property    
    def defaultName(self):
        return ""
    
    @Property
    def richName(self):
        defaultname = self.defaultName
        if defaultname:
            if self.name:
                return "{0}_{1}".format(self.name,defaultname)
            return defaultname
        return self.name

class Leurre(ModelBase):
    def __init__(self, owner=None, model=None):
        # initialisation d'un leurre en copie d'un modele
        super(Leurre, self).__init__(owner, model.name)
        self.trueName = model.trueName
        self.hidden = model.hidden
        self.biasBased = model.getBiasBased()
        self.outputLinks = model.getOutputLinks()
        self.paramCount = model.paramCount
        self.dimension = model.dimension
        self.inputCount = model.inputCount
        self.outputCount = model.outputCount
        self.inputNames = [model.inputNodes[ind].nameC() for ind in range(self.inputCount)]
        self.outputNames = [model.outputNodes[ind].nameC() for ind in range(self.outputCount)]
        try:
            self.isLinear = model.isLinear
        except AttributeError:
            try:
                self.isLinear = model.owner.isLinear
            except:
                self.isLinear = not bool(self.hidden)
        self.smiles = model.smiles
        try:
            self.doubt = model.doubt
        except AttributeError:
            self.doubt = False

@Properties(("originallink", 'useOriginal', 'longParamCount',), False)
class _Model (ModelBase):
    source = None
    def __init__(self, owner=None, name=""):  # Model
        super(Model, self).__init__(owner, name)
        self._comments = []
        self.style = 0
        self._isLinear = None
        self.curinputs = None
        self._truename = ""
        self._paramdict = None
        self._originallink = True
        self._useOriginal = False
        self._longParamCount = 0
        self._modelType = 1
        self._extraInputCount = 0
    
    @Property
    def trueName(self):  # Model
        return self._truename if self._truename else self.name 
    @trueName.setter
    def strueName(self, value):  # Model
        self._truename = value 
    
    def propagate(self, inputs=None, weights=None):  # Model
        return self
        
    def __call__(self, inputs=None, computelevel=C.ID_COMPUTE, gradient=None, 
                 gradInput=None, hessian=None, original=True):  # Model
        inputs = None if inputs is None else asarray(inputs)
        if (inputs is not None) and (isinstance(inputs, IndexedProperty) or (inputs.ndim==2)):
            res = [self.__call__(locinput, computelevel, gradient, gradInput, 
                                 hessian, original) for locinput in inputs]
            return asarray(res)    
        self.computelevel = computelevel 
        self.propagate(inputs)
        if self.computelevel <= C.ID_COMPUTE:
            # pas de demnde de calcul des gradients
            return self.getOutputvalues()
        
        # ici on va calculer les gradients par rapport aux poids et par rapport aux entrées.
        # le supplément pour calcul du 2eme gradient est très faible
        ndata = self.inputDim
        internalgrad = gradient is None
        internalgradIn = (gradInput is None) and (ndata == 1) and self.layerLength(0)
        no = self.outputCount
        if self.useOriginal:
            ns = self.longParamCount
        else:
            ns = self.synapseCount
        
        if no == 1:
            if internalgrad:
                if ns <= 1:
                    gradient = zeros((ndata,))
                else:
                    gradient = zeros((ndata, ns)).squeeze()
            if internalgradIn:
                gradInput = zeros((self.layerLength(0),))
            outGrad = 1
                
            self.backPropagateCoreSingleOutput(gradient, gradInput, outGrad, original=original)
                
            if internalgrad:
                gradient = gradient.squeeze()
            
        else:
            if internalgrad:
                gradient = zeros((no, ndata, ns))
            if internalgradIn:
                gradInput = zeros((no, self.layerLength(0)))
            outGrad = identity(no)
                
            self.backPropagateCoreMultiOutput(gradient, gradInput, outGrad)
                
            if internalgrad:
                gradient = gradient.squeeze()
            if internalgradIn:
                gradInput = gradInput.squeeze()
        
        #if not self.computeLevel:
        #    return self.outputs
        if (gradInput is None) or not len(gradInput): # or not gradInput:
            return self.getOutputvalues(), gradient
        return self.getOutputvalues(), gradient, gradInput
        
    # = transferOld
    
    def getOutputvalues(self):  # Model
        if self.outputCount == 1:
            return self.outputs[0]
        return array([val for val in self.outputs])
    
    @property
    def isLinear(self):
        return self.getIsLinear()
    
    def getIsLinear(self):  # Model
        if self._isLinear is not None:
            return self._isLinear
        hasw = self._weights is not None
        if self._isLinear is None:
            result = True
            if hasw:
                memweight = self._weights.copy()
            #ww = randn(dim,)
            #self.setWeights(ww)
            self.initWeights()
            inputs = randn(self.inputCount,)
            res1 = self(inputs, computelevel=C.ID_DIFF, original=False)
            #ww = randn(dim,)
            #self.setWeights(ww)
            self.initWeights()
            res2 = self(inputs, computelevel=C.ID_DIFF, original=False)
            
            for r1, r2 in zip(res1[1], res2[1]):
#                 for val1, val2 in zip(r1, r2):
                if abs(r1 - r2) > 1E-10:
                    result = False
                    break
#                 if not result:
#                     break
            if hasw:
                self.setWeights(memweight)
            else:
                self._weights = None
            self._isLinear = result
        return self._isLinear
    @isLinear.setter
    def sisLinear(self, value):  # Model
        self._isLinear = value
        
    def signature(self):  # Model
        memweights = self._weights.copy()
        self._weights = [0.1 for _ in range(self.paramCount)]
        inputs = [0.1 for _ in range(self.inputCount)]
        res = self(inputs)
        self._weights = memweights
        return res
        
    def _expandvec(self, vec, inivec, fixed): # copie depuis _driver
        index = 0
        target = []
        for inivalue, fix in zip(inivec, fixed):
            if fix:
                val = inivalue
            else:
                val = vec[index]
                index += 1
            target.append(val)
        return asarray(target)
    
    def _reducevec(self, vec, fixed): # copie depuis _driver
        return asarray([val for test, val in zip(fixed, vec) if not test])

    def reverse_train(self, outputtarget, inputs=None, fixedinputs=None, 
        epochs=None, modelindex=-1, callback=None, debug=0, verbose=0):  # Model
        """Obtain the result of model reverse transfer.
        
        Parameters:
        
         - outputtarget -> output value target.
         - inputs -> initial input values. It may be a vector-like, a dictionary or None:
             - vector-like : vector of the initial input values, in the order of the inputs.
             - dictionary: initial input value for each input name.
             - None : no new initial input values (default).
         - fixedinputs -> fixed input values.  It may be a vector-like, a dictionary or None:
             - vector-like : vector of boolean values, in the order of the inputs.
             - dictionary : each input may be fixed (or not) by its name. A missing input is considered as free.
             - None : all input values are free
         - acc -> accuracy of targetting. Optimizing computation is stopped as soon as this accuracy is obtained, no matter the epochs value.
         - epochs -> maximum number of optimizing computation epochs.
         - callback -> callback function. It takes a current index as parameter.
         - modelindex -> if -1, the computation take into account all the extra weights. Otherwise, if positive, give the index of the model used.
        """
        if not self.inputCount:
            raise MonalError("Model must have at least 1 input to be reverse-trained")
        acc=1.49012e-08
        if fixedinputs is None:
            fixedinputs = [0 for _ in range(self.inputCount)]
        if isinstance(fixedinputs, dict):
            lst = []
            for inname in self.inputNames:
                if inname in fixedinputs:
                    lst.append(fixedinputs[inname])
                else:
                    lst.append(0)
            fixedinputs = lst
        if inputs is None:
            inputs = asarray(self.inputs)
        #inputs = self.normalizeInputs(inputs)
        inputini = [val for val in inputs]
        #outputtarget = self.unnormalizeOutputs(outputtarget, reverse=True)
        args=(inputini, fixedinputs, outputtarget, modelindex)
        factor = self.inputCount    
        reducedinputs = self._reducevec(inputs, fixedinputs)

        newreducedinputs, code = optim_bfgs(self.inputTrainCost, reducedinputs, args, 
            self.inputTrainGrad, maxiter=epochs, disp=verbose >= 3, factor=factor, 
            callback=callback)
        
        if debug:
            if code & 1:
                print_('Information: no acceptable point found', file=sys.stderr)
            if code & 2: 
                print_('Information: accuracy reached', file=sys.stderr)
            if code & 4: 
                print_('Information: stop for maxiter', file=sys.stderr)
        newinputs = self._expandvec(newreducedinputs, inputini, fixedinputs)
        #newinputs = self.normalizeInputs(newinputs, reverse=True)
        return newinputs, code

    def transfer(self, inputs=None, style=C.TFR_STD, original=True, **kwds):  # Model
        # ici transfer sur un modele , pas sur un driver.kwds['index'] n'est pas 
        # pris en compte
        newWeights = kwds.pop('weights', None)
        if newWeights is not None:
            memweights = self._weights
            self._weights = newWeights
        indexmove = kwds.pop('indexmove', -1)
        if (inputs is not None) and (0 <= indexmove < self.inputCount):
            indexinputs = kwds.pop('indexinputs', [])
            result = []
            inival = inputs[indexmove] 
            for value in indexinputs:
                inputs[indexmove] = value
                res = self.transfer(inputs, style=style, original=original)
                result.append(res)
            inputs[indexmove] = inival
        
        elif style == C.TFR_STD:  # 0
            result = self(inputs)  #.transferOld
        elif style == C.TFR_GRADIENT:  # 1
            result = self(inputs, C.ID_DIFF, original=original)  #.transferOld
        elif style == C.TFR_LEVERAGE:  # 2
            res = self(inputs, C.ID_DIFF, original=original)  #.transferOld
            result = res  # a revoir
        elif style == C.TFR_GRADIENTLEVERAGE:  # 3
            res = self(inputs, C.ID_DIFF, original=original)# a revoir  #.transferOld
            pass
            result = res  # a revoir
        elif style == C.TFR_GRADIENTCONFIDENCE:  # 4
            res = self(inputs, C.ID_DIFF, original=original)# a revoir  #.transferOld
            pass
            result = res  # a revoir
        elif style == C.TFR_GRADIENTINPUTS:  # 5
            res = self(inputs, C.ID_DIFF, original=original)  #.transferOld
            result = res[0], res[2]
        elif style == C.TFR_GRADIENTHESSIAN:  # 6
            res = self(inputs, C.ID_DIFF_SEC, original=original)  #.transferOld
        if newWeights is not None:
            self._weights = memweights
        return result  # a revoir
    
    def initWeights(self, stddev=0.3, bias0=False, seed=None): 
        raise Exception('initWeights must be implemented in children classes of Model')
        
    def loadWeightsXML(self, source):  # Model
        return None
        
    def maxParam(self):  # Model
        return -1
    
    def isModel(self):  # Model
        return True
    
    def isNorm(self):  # Model
        return False
    
    def addComment(self, value=""):  # Model
        self._comments.append(value)
        
    def getComment(self, index=0):  # Model
        try: return self._comments[index]
        except: return ""
        
    def getComments(self):  # Model
        return '\n'.join(self._comments)
    
    @Property
    def comments(self, index):  # Model
        return self._comments[index]
    @comments.lengetter
    def lcomments(self):  # Model
        return len(self._comments)
    
    @Property
    def inputs(self, index):  # Model
        raise IndexError
    @inputs.setter
    def inputs(self, index, value):  # Model
        raise IndexError
    @inputs.lengetter
    def inputs(self):  # Model
        return self.inputCount
    
    @Property
    def outputs(self, index):  # Model
        raise IndexError
    @outputs.setter
    def outputs(self, index, value):  # Model
        raise IndexError
    @outputs.lengetter
    def outputs(self):  # Model
        return self.inputCount
    
    @Property
    def commentCount(self):  # Model
        return len(self._comments)
    
    @Property 
    def extendedInputCount(self):
        return self.inputCount + self._extraInputCount        
    
    @Property
    def inputCount(self):  # Model
        return 0
    
    @Property
    def outputCount(self):
        return 0#self._outputcount
    
    @Property
    def order(self):  # Model
        return 0

    @Property
    def hiddenCount(self):  # Model
        return 0
    
    @Property
    def synapseCount(self):  # Model
        return 0
    
#     @Property
#     def weights(self, index):  # Model
#         return []
#     @weights.lengetter
#     def lweights(self):  # Model
#         return 0
#     @weights.setter
#     def sweights(self, index, value):  # Model
#         pass
    
    @Property
    def paramCount(self):  # Model
        return 0
    
    def hasMinMax(self):  # Model
        return False
    
    def saveModel(self, filename, savingformat=C.SF_XML, 
        modeltemplate="", callback=None, compiler=""):  # Model
        if savingformat in [C.SF_XML, C.SF_BINARY, C.SF_JSON]:
            return self.saveObj(self, filename, savingformat)
        else:
            raise Exception('saveModel not implemented yet with format = {}'.format(savingformat))
    
    def saveParameters(self, filename, savingformat=C.SF_XML, count=0, appending=False):  # Model
        if savingformat in [C.SF_XML, C.SF_BINARY, C.SF_JSON]:
            return self.saveObj(self._weights, filename, savingformat, appending)
        else:
            raise Exception('saveModel not implemented yet with format = {}'.format(savingformat))
    
    def saveObj(self, obj, target, savingformat=C.SF_XML, appending=False):  # Model
#        if isinstance(target, basestring):
        if isinstance(target, string_types):
            basedir = os.path.dirname(target)
            safemakedirs(basedir)
            #if not os.path.exists(basedir):
            #    os.ma kedirs(basedir)
            if appending:
                mode = "ab"
            else:
                mode = "wb"
            with open(target, mode) as ff:
                res = self.saveObj(obj, ff, savingformat)
            return res
        if savingformat == C.SF_XML:
            try:
                return target.write(obj.xml().encode())
            except:
                return target.write(obj.xml())
        if savingformat == C.SF_BINARY:
            return pickle.dump(obj, target, pickle.HIGHEST_PROTOCOL)
        if savingformat == C.SF_JSON:
            return json.dump(self)

    @Property 
    def paramdict(self):  # Model
        if self._paramdict is None:
            self._paramdict = self.getParamDict(True)
        return self._paramdict
    
    def getIndexWeightOfName(self, wname):  # Model
        """Retourne l'indice d'un poids de nom défini
        """
        return self.paramdict[wname]
    
    def loadFromASCIIStrList(self, lst, index):  # Model
        res = bool(lst)
        if not res:
            return 0
        dim = self.paramCount
        #i0 = 0
        wformat = []
        #delta = 0
        st= lst[index]
        llst = st.split('=')
        if llst[0].strip().lower() == "count":
            #index += 1
            dim = int(llst[1].strip())
            index += 1
            #st = lst[index]
        for row in range(dim):
            st = lst[index + row]
            locallst = st.split(";")
            if isStrNum(locallst[0]):
                self.weights[row] = floatEx(locallst[0])
                wformat.append(row)
            elif (len(locallst) >= 2) and isStrNum(locallst[1]):
                wname = locallst[0]
                if isinstance(wname, string_types):
                    wname = self.getIndexWeightOfName(wname)
                if wname >= 0:
                    self.weights[wname] = floatEx(locallst[1])
                    wformat.append(wname)
        return dim

# def _decorator(cls, prop_names, readonly):
#     for name in prop_names:
#         hname = '_' + name
#         if not hasattr(cls, hname):
#             setattr(cls, hname, None) 
#         getter = makefct(hname, False)
#         setter = None if readonly else makefct(hname)
#         setattr(cls, name, Property(getter, setter))    
#     return cls        
# 
#         
# Model =  _decorator(_Model, ("originallink", 'useOriginal', 'longParamCount',), False) 
Model = _Model

if __name__ == "__main__":
    pass
#     inidir = "/Users/jeanluc/Desktop/Reserve/Tests SAP Samir MEDROUK"
#     source = os.path.join(inidir, "Modeles fusionnes", "Modele_Final.NML")
#     inifile = os.path.basename(source)
# 
#     with open(source, "r") as ff:
#         sourcecontent = ff.read()
     


#-*- coding: ISO-8859-15 -*-
#===============================================================================
#  Module  monal
#  Projet MonalPy
#
#  Implementation python de monal
#
# Copyright 2016 Jean-Luc PLOIX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================
"""monal module version 3.3
non-linear process modelling.
Sub modules :
    - monalconst   -> common constants
    - monalrecords -> data structures for ctype communicatioon with NDK DLL
    - specialmath  -> special numbers: INF, -INF, NAN
    - util         -> utilities
    - Property     -> special property module, adding formal vector properties feature
    - version      -> module version management
Sub modules for full Python version :
    - model        -> linear and non-linear python models;
    - driver       -> python model driver.
    - lcode        -> C code model
    - include      -> include files for C models
    - library      -> computed libraries management
"""

import os, sys
try:
    __path__
except:
    __path__ = [os.path.dirname(__file__)]
if not __path__[0] in sys.path:
    sys.path.append(__path__[0])

from metaphor.version import __version__
from metaphor.nntoolbox.utils import  getancestor, addpathlist, null
from datetime import datetime

#print("checking 'monal' rights")
CHECK_TOKEN = 1
if CHECK_TOKEN:
    from metaphor.nntoolbox.toktoolbox.tokens import checkToken
#     tokens = __import__("metaphor.nntoolbox.toktoolbox.tokens")
#     checkToken = tokens.checkToken
else:
    checkToken = null
#     from metaphor.nntoolbox.toktoolbox.tokens import checkToken

def version():
    """return the version of the module monal
    """
    return __version__

def installFileName(ppath="dist", ext=".tar.gz"):
    """
    """
    cur = "monal-%s%s" % (__version__, ext)
    if len(ppath):
        return os.path.join(ppath, cur)
    return cur


# try:
#     __path__
# except:
#     __path__ = os.path.dirname(__file__)

try:
    # initialisation de la variable __paths__
    from metaphor.nntoolbox.utils import  addpathlist
    __paths__ = addpathlist(__path__[0], "monal", [])
except:
    __paths__ = __path__

__prefix__ = ""

def prefix():
    global __prefix__
    if not __prefix__:
        __prefix__ = getancestor(__path__[0], 'lib', True, True)
    if not __prefix__:
        __prefix__ = getancestor(__path__[0], 'src', True, True)
    return __prefix__
 
__modulepath__ = ""

def path():
    global __modulepath__
    if not __modulepath__:
        __modulepath__ = __path__[0]
    return __modulepath__
 
applibdir = os.path.join(path(), "lcode")
includedir = os.path.join(path(), "include")
parentdir = os.path.dirname(path())

def tokenPath():
    res = os.path.expanduser("~/.tok")
    if not os.path.exists(res):
        os.makedirs(res)
    return res
tokenfile =  os.path.join(tokenPath(), ".tok.tk")   

def getLocalToken():
    if os.path.exists(tokenfile):
        with open(tokenfile, "r") as ff:
            res = ff.read() 
    else:
        res = ""
    return res

def doCheckToken():
    tok = getLocalToken()
    result = tok
    if not tok:
        tok = input("Please enter a valid user token : ")  #.encode('utf_8')
        result = checkToken(tok, "Monal")
        if not result:
            raise Exception("invalid usertoken")
        tokfile = os.path.join(tokenPath(), ".tok.tk")
        with open(tokfile, "w") as ff:
            ff.write(tok)
    else:
        result = checkToken(tok, "Monal")
    return result

# try:
#     res = doCheckToken()
#     expires = res["expires"]
#     expdate = datetime.fromtimestamp(expires).strftime("%A, %B %d, %Y %I:%M:%S")
#     print("\ntoken expires", expdate)
# except Exception as err:
#     print (err)




"""Package nn1 api
"""
import sys, os
from metaphor.version import __version__

__modulepath__ = ""
def path():
    global __modulepath__
    if not __modulepath__:
        __modulepath__ = __path__[0]
    return __modulepath__

# filelist = []
# for val in sys.path:
#     if val.endswith('metaphor'):
#         for (dirpath, dirnames, filenames) in os.walk(val):
#             filelist.extend([os.path.join(dirpath, val) for val in filenames if val.startswith('services')])
#             break
# servicesList = filelist


# -*- coding: ISO-8859-1 -*-
#-------------------------------------------------------------------------------
# $Id$
#
# Copyright 2016 Jean-Luc PLOIX
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#-------------------------------------------------------------------------------
'''
Created on 20 févr. 2017

@author: jeanluc
'''
import sys
from .run.nn_run import supervisor
supervisor(sys.argv)



#-*- coding: ISO-8859-15 -*-
#===============================================================================
# $Id: _linmodel.py 4819 2018-11-02 05:32:25Z jeanluc $
#  Module  _linmodel
#  Projet MonalPy
# 
#  Implementation python de monal
#
#  Author: Jean-Luc PLOIX  -  NETRAL
#  Decembre 2009
#
# Copyright 2016 Jean-Luc PLOIX
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================
'''
'''

from numpy import matrix, copy, zeros
from ..Property import Property
from ._model import Model
from ..monalconst import ID_COMPUTE
from ..util.monaltoolbox import register

class LinearModel(Model):
    
    def __init__(self, owner=None, name=""):
        super(Model, self).__init__(owner, name)
        self._inputs = None
        self._outputs = None

    def _setNames(self, inNames=None, outNames=None):
        if inNames: 
            self._inputnames = [name for name in inNames]
        else:
            self._inputnames = ["IN_%d"% i for i in range(self.inputCount)]
        if outNames: 
            self._outputnames = [name for name in outNames]
        else:
            self._outputnames = ["OUT_%d"% i for i in range(self.outputCount)]            
    
    def container(self):
        return self.owner
    
    @Property
    def inputs(self, index):
        return self._inputs[index]
    @inputs.setter
    def inputs(self, index, value):
        self._inputs[index] = value
    @inputs.lengetter
    def inputs(self):
        return self.inputCount
    
    @Property
    def outputs(self, index):
        return self._outputs[index]
    @outputs.setter
    def outputs(self, index, value):
        self._outputs[index] = value
    @outputs.lengetter
    def outputs(self):
        return self.inputCount
    
    @Property
    def inputNames(self, index):
        return self._inputnames[index]
    @inputNames.lengetter
    def linputNames(self):
        return self.inputCount
    @inputNames.setter
    def sinputNames(self, index, value):
        self._inputnames[index] = value
    
    @Property
    def outputNames(self, index):
        return self._outputnames[index]
    @outputNames.lengetter
    def loutputNames(self):
        return self.outputCount
    @outputNames.setter
    def soutputNames(self, index, value):
        self._outputnames[index] = value
        

class Normalizer(LinearModel):
    '''
    Normalizing model.
    Data are in a numpy.matrix of sizes (N, 2)
    '''

    def __init__(self, owner=None, data=None, inNames=None, outNames=None, 
                 style=0, name="", reverse=False): 
        super(Normalizer, self).__init__(owner, name)
        self.style = style 
        if (data is None) or isinstance(data, int):
            try: count = int(data)
            except: count = 1
#            data = matrix([[1, 0] for _ in range(count)])  modif le 03/04/2019
            data = zeros((count, 2))
            for ind in range(count):
                data[ind, 0] = 1
        self.setData(data, reverse)
        self._setNames(inNames, outNames)
                        
    def setData(self, data, reverse=False):
        self.data = copy(data)
        if reverse:
            self.data[:, 0] = 1/self.data[:, 0]
            self.data[:, 1] = - self.data[:, 1] * self.data[:, 0]
        self._inputs = [0.0 for _ in range(self.inputCount)]
        self._outputs = [0.0 for _ in range(self.inputCount)]
                
    def isNorm(self):
        return True
        
    @Property
    def inputCount(self):
        try:
            return self.data.shape[0]
        except:
            return 0
    
    @Property
    def outputCount(self):
        try:
            return self.data.shape[0]
        except:
            return 0
    
    def propagate(self, inputs=None):
        if inputs is not None:
            self._outputs = inputs * self.data[:, 0] + self.data[:, 1]            
        else:
            self._outputs = self._inputs * self.data[:, 0] + self.data[:, 1]
        return self
    
    def backPropagate(self, gradient=None, gradOutput=0, computelevel=ID_COMPUTE):
        """Retropropagation dans le modèle.
        
           gradOutput est le vecteur directeur de la retropropagation.
           Retourne un tuple constitue de :
               le vecteur gradient par rapport aux poids = None
               le vecteur gradient par rapport aux entrées
        """  
        return gradient, gradOutput * self.data[:, 0]

#    def inputNormalization(self, index):
#        return self.data[index, :]
#
#    def outputNormalization(self, index):
#        return self.data[index, :]

class Polynom(LinearModel):
    '''
    # under development
    '''
    
    def __init__(self, owner=None, inputs=0, output=0, order=1, inNames=None, outNames=None, name="model_0"):
        super(Polynom, self).__init__(owner, inNames, outNames, name)
        # under development
          
    def propagate(self, inputs=None, weights=None):
        # under development
        return self

    def backPropagate(self, gradient, gradInput=None, gradOutput=None, computelevel=ID_COMPUTE): 
        # under development 
        return None, None

register(Normalizer)       
register(Normalizer, "NORMALIZER") 
register(Polynom)       
register(Polynom, "POLYNOM") 



# -*- coding: ISO-8859-1 -*-
#-------------------------------------------------------------------------------
# $Id$
#
# Copyright 2016 Jean-Luc PLOIX
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#-------------------------------------------------------------------------------
'''
Created on 10 juin 2016

@author: jeanluc
'''

import sys, os
import numpy as np
import random
from ctypes import byref, c_int, c_void_p

from .. import monalconst as C
from ...nntoolbox.utils import RE_LOO, RE_SUPER, BSGROUP 
from ...nntoolbox.sqlite_numpy import insertDataToDb

DEBUG_TRAIN_JOB = 0

no_parallel_seed = os.environ.get('NO_PARALLEL_SEED', 0)

def onEndIter(history, trainend, index, counter, epochs, progress=None, 
              dowrite=sys.stdout.write):
    
    message = ""
    curindex = index+1
    docontinue = (index >= 0)
    if docontinue and progress:
        docontinue = progress(counter+1)
        if isinstance(docontinue, tuple):
            docontinue = docontinue[0]
    if dowrite:
        if not docontinue:
            dowrite("Training aborted", color='green')
            dowrite("\n", color='black')
        elif history is not None:    
            lst = ["" if (index < 0) else "training #%s"% (index+1)]
            lst.append( "\tepoch: %s"%(epochs))
            lst.append("\tini-stddev: %s"%(history[0]))
            lst.append("\tend-stddev: %s"%(history[epochs]))
            if trainend:
                lst.append("\ttraining end: %s"% C.END_TRAINING_DICT[trainend])
            lst.append("")
            message = "\n".join(lst)
            dowrite(message)
        
    return not docontinue #, message

def onSpecialResult(dbConnect, tableTmpl, code, epochs, startweights, 
        endweights, molindex, initindex, indexReturn, cost, press, out, lev, 
        resultvector, dispersion, inittype, target, locseed):  #
    startweights0 = 0.0
    for value in startweights:
        if value != 0:
            startweights0 = value
            break 
    disp0 = dispersion[0,0] if dispersion is not None else None
    tablename = tableTmpl % molindex #"trainingResults%03d"% molindex
    data = {cname: val for cname, val in zip(C.CRITERION_NAME.values(), 
                                             resultvector)}
    data.update({
        'param_start': startweights,
        'param_start0': startweights0,
        'param_end': endweights,
        'param_end0': endweights[0],
        'dispersion': dispersion,
        'dispersion0': disp0,
        'code': code,
        'indtrain': initindex,
        'out': out,
        'leverages': lev,
        'inittype': C.INIT_NAME[inittype],
        'seed': locseed,
        'target': target
        })
    insertDataToDb(dbConnect, table=tablename, data=data)
    return 0

def onReturnResult(dbConnect, tablename, indcur, resultvector, startweights, 
                   endweights, leverages, dispersion, history, locseed):
    epochs = int(resultvector[C.TCR_EPOCHS])
    disp0 = dispersion[0,0] if dispersion is not None else None
    data = {cname: C.CRITERION_TYPE[tcr](val) for (tcr, cname), 
            val in zip(C.CRITERION_NAME.items(), resultvector)}
    startweights0 = 0.0
    for value in startweights:
        if value != 0:
            startweights0 = value
            break 
    data.update({'ID': indcur+1,
                 'traintype': 0,
                 'ini_stddev': history[0],
                 'end_stddev': history[epochs],
                 'param_start': startweights,
                 'param_start0': startweights0,
                 'param_end': endweights,
                 'param_end0': endweights[0],
                 'leverages': leverages,
                 'leverages0': leverages[0],
                 'dispersion': dispersion,
                 'dispersion0': disp0,
                 'seed': locseed})
    insertDataToDb(dbConnect, table=tablename, data=data)

def trainJob(model, weightsstddev, epochs, withhistory, index, seed):
    assert epochs
    if no_parallel_seed:
        locseed = seed
        locseed0 = seed
    elif (seed is not None) and (seed >= 0):
        locseed0 = seed + (index+1) * 0x1000   
        # 21/03/2019 locseed = seed + index * 0x1000
        random.seed(locseed0)  # 18/04/2018
        locseed = random.randint(0x10, 0xFFFFFFFF)  # 18/04/2018
    else:
        locseed0 = None
        locseed = None
    model.setRandSeed(locseed)
    code = 0
    if not index:
        weights = model.weights.copy()
    else:
        weights = model.newWeights(weightsstddev)
    memweights = weights.copy()
    try:
        res = model.train_integrated(weights, epochs, withhistory, 
                                     computeDispersion=True)  # True
        cur, N, stddev, trainend, hist, code = res[0] 
        if withhistory:
            history = np.sqrt(hist / model.dataCount)
        else:
            history = None  
    except Exception as exres:
        res = [None, None, None]
        N = -1
        stddev = 0.0
        cur = None
        print("trainJob error:", exres)
        if not code:
            code = -1
        res0 = (index, cur, N, stddev, None, None,  None, locseed, code)      
        return res0, res[1], res[2]
    if code:
        cur = None
        stddev = 0.0
        
    res0 = (index, cur, N, stddev, memweights, trainend,  history, locseed0, 
            code)
    return res0, res[1], res[2]

def trainJobBS(model, initlist, inittype, initindex, residuals, BStype,
        settrainindex, epochs, index, computeDispersion, weightsstddev, 
        targets, seed):    
    model.settrainingset()
    epochsmem = epochs    
    BSindex = index if BStype == RE_LOO else BStype 
     
    if no_parallel_seed:
        locseed = seed  
        locseed0 = seed
    elif (seed is not None) and (seed >= 0):
        locseed0 = seed + ((index + 1) * 0xFFFF) + (initindex * 2)  # 0xFFFF = 65535
        random.seed(locseed0)
        locseed = random.randint(0x10, 0xFFFFFFFF)
    else:
        locseed = None
        
    model.setRandSeed(locseed)
    code = model.settrainingset(settrainindex, BSindex, residuals)
    assert (code == 0)
    if inittype == C.INIT_START_PARAM:
        initrecord = initlist[initindex]
        params = initrecord[2].copy()
    elif inittype == C.INIT_END_PARAM:
        initrecord = initlist[initindex]
        params = initrecord[3].copy()
    else:
        params = model.newWeights(weightsstddev)
    if BStype == RE_SUPER:
        indexReturn = initrecord[0]    
    else:
        indexReturn = initindex

    ini_params = params.copy()
    
    res = model.train_integrated(params, epochs=epochsmem, withhistory=False, 
                                 computeDispersion=computeDispersion)
    params, epochs, stddev, trainend, hist, code = res[0]
    press, residuals, singular, leverages, disp, svdlen, deltash, outputs, \
    jacobians = res[1]
    resvec = res[2]
    if BStype == RE_LOO:
        target = targets[index]
        if computeDispersion:
            out, gradient = model.transfergradientindex(params, index=index)
            lev = np.dot(gradient.T, np.dot(disp, gradient))  
    elif BStype in BSGROUP:
        target = None
        out = None
        lev = None
    else:
        out = model.transferindex(params, index=index)
        lev = 0.0
        target = targets[index]
    return  code, epochs, ini_params.copy(), params.copy(), index, initindex, \
        indexReturn, stddev, press, out, lev, resvec, disp, inittype , target, \
        locseed0                                                      
        



#-*- coding: ISO-8859-15 -*-
#===============================================================================
#  Module  driver
#  Projet MonalDLL
# 
#  Implementation python de monal
#
#  Author: Jean-Luc PLOIX  -  NETRAL
#  Decembre 2009
#
# Copyright 2016 Jean-Luc PLOIX
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================
"""module __init__.
A developper
"""
from metaphor.nntoolbox.dataset import DataManager
from ._driver import DriverBase, Driver, saveModel, setCallback, \
        setSavingFormat
from ._optimizer import optim_bfgs, optim_leastsq
from ._workshop import Workshop
from ._networkfactory import NetworkFactory
from ._drivermultidyn import DriverMultiDyn #, doBootstrap, multitrain
from ._parallelJobs import trainJobBS, trainJob, onEndIter, onReturnResult, \
        onSpecialResult
from ._driverlib import DriverLib, DriverLibError

__all__ = ["DataManager", "Driver", "saveModel", "Workshop",
           "optim_bfgs", "optim_leastsq", "NetworkFactory",
           "setCallback", "setSavingFormat", "DriverMultiDyn",
           "DriverBase", "DriverMultiDyn", "trainJob", "trainJobBS",
           "onEndIter", "onReturnResult", "onSpecialResult", "DriverLib",
           "DriverLibError"]  #, "setUnicode"